# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum
from .LibraryUtility import AppStoreServerLibraryEnumMeta

class TransactionReason(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The cause of a purchase transaction, which indicates whether it’s a customer’s purchase or a renewal for an auto-renewable subscription that the system initiates.
    
    https://developer.apple.com/documentation/appstoreserverapi/transactionreason
    """
    PURCHASE = "PURCHASE"
    RENEWAL = "RENEWAL"



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from typing import List
from base64 import b64decode
from enum import IntEnum
import time
import datetime

import asn1
import jwt
import requests
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
from cryptography.hazmat.primitives.hashes import SHA1, SHA256
from cryptography.x509 import ocsp, oid
from OpenSSL import crypto

from appstoreserverlibrary.models.AppTransaction import AppTransaction
from appstoreserverlibrary.models.LibraryUtility import _get_cattrs_converter

from .models.Environment import Environment
from .models.ResponseBodyV2DecodedPayload import ResponseBodyV2DecodedPayload
from .models.JWSTransactionDecodedPayload import JWSTransactionDecodedPayload
from .models.JWSRenewalInfoDecodedPayload import JWSRenewalInfoDecodedPayload

class SignedDataVerifier:
    """
    A class providing utility methods for verifying and decoding App Store signed data.
    """
    def __init__(
        self,
        root_certificates: List[bytes],
        enable_online_checks: bool,
        environment: Environment,
        bundle_id: str,
        app_apple_id: int = None,
    ):
        self._chain_verifier = _ChainVerifier(root_certificates)
        self._environment = environment
        self._bundle_id = bundle_id
        self._app_apple_id = app_apple_id
        self._enable_online_checks = enable_online_checks

    def verify_and_decode_renewal_info(self, signed_renewal_info: str) -> JWSRenewalInfoDecodedPayload:
        """
        Verifies and decodes a signedRenewalInfo obtained from the App Store Server API, an App Store Server Notification, or from a device

        :param signed_renewal_info: The signedRenewalInfo field
        :return: The decoded renewal info after verification
        :throws VerificationException: Thrown if the data could not be verified
        """
        
        decoded_renewal_info = _get_cattrs_converter(JWSRenewalInfoDecodedPayload).structure(self._decode_signed_object(signed_renewal_info), JWSRenewalInfoDecodedPayload)
        if decoded_renewal_info.environment != self._environment:
            raise VerificationException(VerificationStatus.INVALID_ENVIRONMENT)
        return decoded_renewal_info

    def verify_and_decode_signed_transaction(self, signed_transaction: str) -> JWSTransactionDecodedPayload:
        """
        Verifies and decodes a signedTransaction obtained from the App Store Server API, an App Store Server Notification, or from a device

        :param signed_transaction: The signedRenewalInfo field
        :return: The decoded transaction info after verification
        :throws VerificationException: Thrown if the data could not be verified
        """
        decoded_transaction_info = _get_cattrs_converter(JWSTransactionDecodedPayload).structure(self._decode_signed_object(signed_transaction), JWSTransactionDecodedPayload)
        if decoded_transaction_info.bundleId != self._bundle_id:
            raise VerificationException(VerificationStatus.INVALID_APP_IDENTIFIER)
        if decoded_transaction_info.environment != self._environment:
            raise VerificationException(VerificationStatus.INVALID_ENVIRONMENT)
        return decoded_transaction_info

    def verify_and_decode_notification(self, signed_payload: str) -> ResponseBodyV2DecodedPayload:
        """
        Verifies and decodes an App Store Server Notification signedPayload

        :param signedPayload: The payload received by your server
        :return: The decoded payload after verification
        :throws VerificationException: Thrown if the data could not be verified
        """
        decoded_dict = self._decode_signed_object(signed_payload)
        decoded_signed_notification = _get_cattrs_converter(ResponseBodyV2DecodedPayload).structure(decoded_dict, ResponseBodyV2DecodedPayload)
        bundle_id = None
        app_apple_id = None
        environment = None
        if decoded_signed_notification.data:
            bundle_id = decoded_signed_notification.data.bundleId
            app_apple_id = decoded_signed_notification.data.appAppleId
            environment = decoded_signed_notification.data.environment
        elif decoded_signed_notification.summary:
            bundle_id = decoded_signed_notification.summary.bundleId
            app_apple_id = decoded_signed_notification.summary.appAppleId
            environment = decoded_signed_notification.summary.environment
        if bundle_id != self._bundle_id or (self._environment == Environment.PRODUCTION and app_apple_id != self._app_apple_id):
            raise VerificationException(VerificationStatus.INVALID_APP_IDENTIFIER)
        if environment != self._environment:
            raise VerificationException(VerificationStatus.INVALID_ENVIRONMENT)
        return decoded_signed_notification

    def verify_and_decode_app_transaction(self, signed_app_transaction: str) -> AppTransaction:
        """
        Verifies and decodes a signed AppTransaction

        :param signed_app_transaction: The signed AppTransaction
        :return: The decoded AppTransaction after validation
        :throws VerificationException: Thrown if the data could not be verified
        """
        decoded_dict = self._decode_signed_object(signed_app_transaction)
        decoded_app_transaction = _get_cattrs_converter(AppTransaction).structure(decoded_dict, AppTransaction)
        environment = decoded_app_transaction.receiptType
        if decoded_app_transaction.bundleId != self._bundle_id or (self._environment == Environment.PRODUCTION and decoded_app_transaction.appAppleId != self._app_apple_id):
            raise VerificationException(VerificationStatus.INVALID_APP_IDENTIFIER)
        if environment != self._environment:
            raise VerificationException(VerificationStatus.INVALID_ENVIRONMENT)
        return decoded_app_transaction

    def _decode_signed_object(self, signed_obj: str) -> dict:
        try:
            decoded_jwt = jwt.decode(signed_obj, options={"verify_signature": False})
            if self._environment == Environment.XCODE or self._environment == Environment.LOCAL_TESTING:
                # Data is not signed by the App Store, and verification should be skipped
                # The environment MUST be checked in the public method calling this
                return decoded_jwt
            unverified_headers: dict = jwt.get_unverified_header(signed_obj)
            x5c_header: List[str] = unverified_headers.get("x5c")
            if x5c_header is None or len(x5c_header) == 0:
                raise Exception("x5c claim was empty")
            algorithm_header: str = unverified_headers.get("alg")
            if algorithm_header is None or "ES256" != algorithm_header:
                raise Exception("Algorithm was not ES256")
            signed_date = decoded_jwt.get('signedDate') if decoded_jwt.get('signedDate') is not None else decoded_jwt.get('receiptCreationDate')
            effective_date = time.time() if self._enable_online_checks or signed_date is None else int(signed_date) // 1000
            signing_key = self._chain_verifier.verify_chain(x5c_header, self._enable_online_checks, effective_date)
            return jwt.decode(signed_obj, signing_key, algorithms=["ES256"])
        except VerificationException as e:
            raise e
        except Exception as e:
            raise VerificationException(VerificationStatus.VERIFICATION_FAILURE) from e

class _ChainVerifier:
    def __init__(self, root_certificates: List[bytes], enable_strict_checks=True):
        self.enable_strict_checks = enable_strict_checks
        self.root_certificates = root_certificates

    def verify_chain(self, certificates: List[str], perform_online_checks: bool, effective_date: int) -> str:
        if len(self.root_certificates) == 0:
            raise VerificationException(VerificationStatus.INVALID_CERTIFICATE)
        if len(certificates) != 3:
            raise VerificationException(VerificationStatus.INVALID_CHAIN_LENGTH)
        trusted_store = crypto.X509Store()
        try:
            for trusted_cert_bytes in self.root_certificates:
                trusted_cert = crypto.load_certificate(crypto.FILETYPE_ASN1, trusted_cert_bytes)
                trusted_store.add_cert(trusted_cert)
            if self.enable_strict_checks:
                trusted_store.set_flags(crypto.X509StoreFlags.X509_STRICT)
            leaf_cert = crypto.load_certificate(crypto.FILETYPE_ASN1, b64decode(certificates[0], validate=True))
            intermediate_cert = crypto.load_certificate(crypto.FILETYPE_ASN1, b64decode(certificates[1], validate=True))
            verification_context = crypto.X509StoreContext(trusted_store, leaf_cert, [intermediate_cert])
        except Exception as e:
            raise VerificationException(VerificationStatus.INVALID_CERTIFICATE) from e

        trusted_store.set_time(datetime.datetime.fromtimestamp(effective_date, tz=datetime.timezone.utc))
        try:
            verification_context.verify_certificate()
            trusted_chain = verification_context.get_verified_chain()
        except Exception as e:
            raise VerificationException(VerificationStatus.VERIFICATION_FAILURE) from e
        self.check_oid(trusted_chain[0].to_cryptography(), "1.2.840.113635.100.6.11.1")
        self.check_oid(trusted_chain[1].to_cryptography(), "1.2.840.113635.100.6.2.1")
        if perform_online_checks:
            self.check_ocsp_status(trusted_chain[1], trusted_chain[2], trusted_chain[2])
            self.check_ocsp_status(trusted_chain[0], trusted_chain[1], trusted_chain[2])
        return (
            leaf_cert.to_cryptography()
            .public_key()
            .public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)
            .decode()
        )

    def check_oid(self, cert: x509.Certificate, oid: str):
        try:
            cert.extensions.get_extension_for_oid(x509.ObjectIdentifier(oid))
        except Exception as e:
            raise VerificationException(VerificationStatus.VERIFICATION_FAILURE) from e

    def check_ocsp_status(self, cert: crypto.X509, issuer: crypto.X509, root: crypto.X509):
        builder = ocsp.OCSPRequestBuilder()
        builder = builder.add_certificate(cert.to_cryptography(), issuer.to_cryptography(), SHA256())
        req = builder.build()
        authority_values = (
            cert.to_cryptography()
            .extensions.get_extension_for_oid(x509.oid.ExtensionOID.AUTHORITY_INFORMATION_ACCESS)
            .value
        )
        ocsps = [val for val in authority_values if val.access_method == x509.oid.AuthorityInformationAccessOID.OCSP]
        for o in ocsps:
            r = requests.post(
                o.access_location.value,
                headers={"Content-Type": "application/ocsp-request"},
                data=req.public_bytes(serialization.Encoding.DER),
            )
            if r.status_code == 200:
                ocsp_resp = ocsp.load_der_ocsp_response(r.content)
                if ocsp_resp.response_status == ocsp.OCSPResponseStatus.SUCCESSFUL:
                    certs = [issuer]
                    for ocsp_cert in ocsp_resp.certificates:
                        certs.append(crypto.X509.from_cryptography(ocsp_cert))
                    # Find signing cert
                    signing_cert = None
                    for potential_signing_cert in certs:
                        if ocsp_resp.responder_key_hash:
                            subject_public_key_info = (
                                potential_signing_cert.get_pubkey()
                                .to_cryptography_key()
                                .public_bytes(
                                    encoding=serialization.Encoding.DER,
                                    format=serialization.PublicFormat.SubjectPublicKeyInfo,
                                )
                            )
                            decoder = asn1.Decoder()
                            decoder.start(subject_public_key_info)
                            decoder.enter()
                            decoder.read()
                            _, value = decoder.read()
                            digest = hashes.Hash(SHA1())
                            digest.update(value)
                            if digest.finalize() == ocsp_resp.responder_key_hash:
                                signing_cert = potential_signing_cert
                                break

                        elif ocsp_resp.responder_name:
                            if ocsp_resp.responder_name == potential_signing_cert.subject.rfc4514_string():
                                signing_cert = potential_signing_cert
                                break
                    if signing_cert is None:
                        raise VerificationException(VerificationStatus.VERIFICATION_FAILURE)

                    if signing_cert.to_cryptography().public_bytes(
                        encoding=serialization.Encoding.DER
                    ) == issuer.to_cryptography().public_bytes(encoding=serialization.Encoding.DER):
                        # We trust this because it is the issuer
                        pass
                    else:
                        trusted_store = crypto.X509Store()
                        trusted_store.add_cert(issuer)
                        trusted_store.add_cert(root)  # Apparently a full chain is always needed
                        verification_context = crypto.X509StoreContext(trusted_store, signing_cert, [])
                        verification_context.verify_certificate()
                        if (
                            oid.ExtendedKeyUsageOID.OCSP_SIGNING
                            not in signing_cert.to_cryptography()
                            .extensions.get_extension_for_class(x509.ExtendedKeyUsage)
                            .value._usages
                        ):
                            raise VerificationException(VerificationStatus.VERIFICATION_FAILURE)

                    # Confirm response is signed by signing_certificate
                    signing_cert.to_cryptography().public_key().verify(
                        ocsp_resp.signature, ocsp_resp.tbs_response_bytes, ECDSA(ocsp_resp.signature_hash_algorithm)
                    )

                    # Get the CertId
                    for single_response in ocsp_resp.responses:
                        # Get the cert ID with the provided hashing algorithm (using the request builder wrapper)
                        builder = ocsp.OCSPRequestBuilder()
                        builder = builder.add_certificate(
                            cert.to_cryptography(), issuer.to_cryptography(), single_response.hash_algorithm
                        )
                        req = builder.build()
                        if (
                            single_response.certificate_status == ocsp.OCSPCertStatus.GOOD
                            and single_response.serial_number == req.serial_number
                            and single_response.issuer_key_hash == req.issuer_key_hash
                            and single_response.issuer_name_hash == req.issuer_name_hash
                        ):
                            # Success
                            return

        raise VerificationException(VerificationStatus.VERIFICATION_FAILURE)


class VerificationStatus(IntEnum):
    OK = 0
    VERIFICATION_FAILURE = 1
    INVALID_APP_IDENTIFIER = 2
    INVALID_CERTIFICATE = 3
    INVALID_CHAIN_LENGTH = 4
    INVALID_CHAIN = 5
    INVALID_ENVIRONMENT = 6


class VerificationException(Exception):
    def __init__(self, status: VerificationStatus):
        super().__init__("Verification failed with status " + status.name)
        self.status = status



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional, List

from attr import define
import attr

from .NotificationHistoryResponseItem import NotificationHistoryResponseItem

@define
class NotificationHistoryResponse: 
    """
    A response that contains the App Store Server Notifications history for your app.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationhistoryresponse
    """

    paginationToken: Optional[str] = attr.ib(default=None)
    """
    A pagination token that you return to the endpoint on a subsequent call to receive the next set of results.
    
    https://developer.apple.com/documentation/appstoreserverapi/paginationtoken
    """

    hasMore: Optional[bool] = attr.ib(default=None)
    """
    A Boolean value indicating whether the App Store has more transaction data.
    
    https://developer.apple.com/documentation/appstoreserverapi/hasmore
    """

    notificationHistory: Optional[List[NotificationHistoryResponseItem]] = attr.ib(default=None)
    """
    An array of App Store server notification history records.
    
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

from .LibraryUtility import AttrsRawValueAware

from .Environment import Environment

@define
class AppTransaction(AttrsRawValueAware):
    """
    Information that represents the customer’s purchase of the app, cryptographically signed by the App Store.
    
    https://developer.apple.com/documentation/storekit/apptransaction
    """

    receiptType: Optional[Environment] = Environment.create_main_attr('rawReceiptType')
    """
    The server environment that signs the app transaction.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3963901-environment
    """

    rawReceiptType: Optional[str] = Environment.create_raw_attr('receiptType')
    """
    See receiptType
    """

    appAppleId: Optional[int] = attr.ib(default=None)
    """
    The unique identifier the App Store uses to identify the app.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954436-appid
    """

    bundleId: Optional[str] = attr.ib(default=None)
    """
    The bundle identifier that the app transaction applies to.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954439-bundleid
    """

    applicationVersion: Optional[str] = attr.ib(default=None)
    """
    The app version that the app transaction applies to.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954437-appversion
    """

    versionExternalIdentifier: Optional[int] = attr.ib(default=None)
    """
    The version external identifier of the app
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954438-appversionid
    """

    receiptCreationDate: Optional[int] = attr.ib(default=None)
    """
    The date that the App Store signed the JWS app transaction.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954449-signeddate
    """

    originalPurchaseDate: Optional[int] = attr.ib(default=None)
    """
    The date the user originally purchased the app from the App Store.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954448-originalpurchasedate
    """

    originalApplicationVersion: Optional[str] = attr.ib(default=None)
    """
    The app version that the user originally purchased from the App Store.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954447-originalappversion
    """

    deviceVerification: Optional[str] = attr.ib(default=None)
    """
    The Base64 device verification value to use to verify whether the app transaction belongs to the device.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954441-deviceverification
    """

    deviceVerificationNonce: Optional[str] = attr.ib(default=None)
    """
    The UUID used to compute the device verification value.
    
    https://developer.apple.com/documentation/storekit/apptransaction/3954442-deviceverificationnonce
    """

    preorderDate: Optional[int] = attr.ib(default=None)
    """
    The date the customer placed an order for the app before it's available in the App Store.
    
    https://developer.apple.com/documentation/storekit/apptransaction/4013175-preorderdate
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import EnumMeta
from typing import Any, List, Type, TypeVar

from attr import Attribute, has, ib, fields
from cattr import override
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override
import cattrs

T = TypeVar('T')

metadata_key = 'correspondingFieldName'
metadata_type_key = 'typeOfField'

class AppStoreServerLibraryEnumMeta(EnumMeta):
    def __contains__(c, val):
        try:
            c(val)
        except ValueError:
            return False
        return True  
    
    def create_main_attr(c, raw_field_name: str) -> Any:
        def value_set(self, _: Attribute, value: c):
            newValue = value.value if value is not None else None
            if newValue != getattr(self, raw_field_name):
                object.__setattr__(self, raw_field_name, newValue)
            return value
        return ib(default=None, on_setattr=value_set, metadata={metadata_key: raw_field_name, metadata_type_key: 'main'})
    
    def create_raw_attr(c, field_name: str) -> Any:
        def value_set(self, _: Attribute, value: str):
            newValue = c(value) if value in c else None
            if newValue != getattr(self, field_name):
                object.__setattr__(self, field_name, newValue)
            return value
        return ib(default=None, kw_only=True, on_setattr=value_set, metadata={metadata_key: field_name, metadata_type_key: 'raw'})
    
class AttrsRawValueAware:
    def __attrs_post_init__(self):
        attr_fields: List[Attribute] = fields(type(self))
        for attribute in attr_fields:
            if metadata_type_key not in attribute.metadata or attribute.metadata[metadata_type_key] != 'raw':
                continue
            field: str = attribute.metadata.get(metadata_key)
            rawField = 'raw' + field[0].upper() + field[1:]
            rawValue = getattr(self, rawField)
            value = getattr(self, field)
            if rawValue is not None:
                setattr(self, rawField, rawValue)
            elif value is not None:
                setattr(self, field, value)

    
def _get_cattrs_converter(destination_class: Type[T]) -> cattrs.Converter:
    c = cattrs.Converter()
    attributes: List[Attribute] = fields(destination_class)
    cattrs_overrides = {}
    for attribute in attributes:
        if metadata_type_key in attribute.metadata:
            matching_name: str = attribute.metadata[metadata_key]
            if attribute.metadata[metadata_type_key] == 'raw':
                cattrs_overrides[matching_name] = override(omit=True)
                raw_field = 'raw' + matching_name[0].upper() + matching_name[1:]
                cattrs_overrides[raw_field] = override(rename=matching_name)
    c.register_structure_hook_factory(has, lambda cl: make_dict_structure_fn(cl, c, **cattrs_overrides))
    c.register_unstructure_hook_factory(has, lambda cl: make_dict_unstructure_fn(cl, c, **cattrs_overrides))
    return c


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class RevocationReason(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The reason for a refunded transaction.
    
    https://developer.apple.com/documentation/appstoreserverapi/revocationreason
    """
    REFUNDED_DUE_TO_ISSUE = 1
    REFUNDED_FOR_OTHER_REASON = 0



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

import unittest
from appstoreserverlibrary.models.AutoRenewStatus import AutoRenewStatus
from appstoreserverlibrary.models.Environment import Environment
from appstoreserverlibrary.models.ExpirationIntent import ExpirationIntent
from appstoreserverlibrary.models.InAppOwnershipType import InAppOwnershipType
from appstoreserverlibrary.models.NotificationTypeV2 import NotificationTypeV2
from appstoreserverlibrary.models.OfferDiscountType import OfferDiscountType
from appstoreserverlibrary.models.OfferType import OfferType
from appstoreserverlibrary.models.PriceIncreaseStatus import PriceIncreaseStatus
from appstoreserverlibrary.models.RevocationReason import RevocationReason
from appstoreserverlibrary.models.Status import Status
from appstoreserverlibrary.models.Subtype import Subtype
from appstoreserverlibrary.models.TransactionReason import TransactionReason
from appstoreserverlibrary.models.Type import Type

from tests.util import create_signed_data_from_json, get_default_signed_data_verifier

class DecodedPayloads(unittest.TestCase):
    def test_app_transaction_decoding(self):
        signed_app_transaction = create_signed_data_from_json('tests/resources/models/appTransaction.json')
        
        signed_data_verifier = get_default_signed_data_verifier()

        app_transaction = signed_data_verifier.verify_and_decode_app_transaction(signed_app_transaction)
        self.assertEqual(Environment.LOCAL_TESTING, app_transaction.receiptType)
        self.assertEqual("LocalTesting", app_transaction.rawReceiptType)
        self.assertEqual(531412, app_transaction.appAppleId)
        self.assertEqual("com.example", app_transaction.bundleId)
        self.assertEqual("1.2.3", app_transaction.applicationVersion)
        self.assertEqual(512, app_transaction.versionExternalIdentifier)
        self.assertEqual(1698148900000, app_transaction.receiptCreationDate)
        self.assertEqual(1698148800000, app_transaction.originalPurchaseDate)
        self.assertEqual("1.1.2", app_transaction.originalApplicationVersion)
        self.assertEqual("device_verification_value", app_transaction.deviceVerification)
        self.assertEqual("48ccfa42-7431-4f22-9908-7e88983e105a", app_transaction.deviceVerificationNonce)
        self.assertEqual(1698148700000, app_transaction.preorderDate)

    def test_transaction_decoding(self):
        signed_transaction = create_signed_data_from_json('tests/resources/models/signedTransaction.json')
        
        signed_data_verifier = get_default_signed_data_verifier()

        transaction = signed_data_verifier.verify_and_decode_signed_transaction(signed_transaction)
        
        self.assertEqual("12345", transaction.originalTransactionId)
        self.assertEqual("23456", transaction.transactionId)
        self.assertEqual("34343", transaction.webOrderLineItemId)
        self.assertEqual("com.example", transaction.bundleId)
        self.assertEqual("com.example.product", transaction.productId)
        self.assertEqual("55555", transaction.subscriptionGroupIdentifier)
        self.assertEqual(1698148800000, transaction.originalPurchaseDate)
        self.assertEqual(1698148900000, transaction.purchaseDate)
        self.assertEqual(1698148950000, transaction.revocationDate)
        self.assertEqual(1698149000000, transaction.expiresDate)
        self.assertEqual(1, transaction.quantity)
        self.assertEqual(Type.AUTO_RENEWABLE_SUBSCRIPTION, transaction.type)
        self.assertEqual("Auto-Renewable Subscription", transaction.rawType)
        self.assertEqual("7e3fb20b-4cdb-47cc-936d-99d65f608138", transaction.appAccountToken)
        self.assertEqual(InAppOwnershipType.PURCHASED, transaction.inAppOwnershipType)
        self.assertEqual("PURCHASED", transaction.rawInAppOwnershipType)
        self.assertEqual(1698148900000, transaction.signedDate)
        self.assertEqual(RevocationReason.REFUNDED_DUE_TO_ISSUE, transaction.revocationReason)
        self.assertEqual(1, transaction.rawRevocationReason)
        self.assertEqual("abc.123", transaction.offerIdentifier)
        self.assertTrue(transaction.isUpgraded)
        self.assertEqual(OfferType.INTRODUCTORY_OFFER, transaction.offerType)
        self.assertEqual(1, transaction.rawOfferType)
        self.assertEqual("USA", transaction.storefront)
        self.assertEqual("143441", transaction.storefrontId)
        self.assertEqual(TransactionReason.PURCHASE, transaction.transactionReason)
        self.assertEqual("PURCHASE", transaction.rawTransactionReason)
        self.assertEqual(Environment.LOCAL_TESTING, transaction.environment)
        self.assertEqual("LocalTesting", transaction.rawEnvironment)
        self.assertEqual(10990, transaction.price)
        self.assertEqual("USD", transaction.currency)
        self.assertEqual(OfferDiscountType.PAY_AS_YOU_GO, transaction.offerDiscountType)
        self.assertEqual("PAY_AS_YOU_GO", transaction.rawOfferDiscountType)

    
    def test_renewal_info_decoding(self):
        signed_renewal_info = create_signed_data_from_json('tests/resources/models/signedRenewalInfo.json')
        
        signed_data_verifier = get_default_signed_data_verifier()

        renewal_info = signed_data_verifier.verify_and_decode_renewal_info(signed_renewal_info)

        self.assertEqual(ExpirationIntent.CUSTOMER_CANCELLED, renewal_info.expirationIntent)
        self.assertEqual(1, renewal_info.rawExpirationIntent)
        self.assertEqual("12345", renewal_info.originalTransactionId)
        self.assertEqual("com.example.product.2", renewal_info.autoRenewProductId)
        self.assertEqual("com.example.product", renewal_info.productId)
        self.assertEqual(AutoRenewStatus.ON, renewal_info.autoRenewStatus)
        self.assertEqual(1, renewal_info.rawAutoRenewStatus)
        self.assertTrue(renewal_info.isInBillingRetryPeriod)
        self.assertEqual(PriceIncreaseStatus.CUSTOMER_HAS_NOT_RESPONDED, renewal_info.priceIncreaseStatus)
        self.assertEqual(0, renewal_info.rawPriceIncreaseStatus)
        self.assertEqual(1698148900000, renewal_info.gracePeriodExpiresDate)
        self.assertEqual(OfferType.PROMOTIONAL_OFFER, renewal_info.offerType)
        self.assertEqual(2, renewal_info.rawOfferType)
        self.assertEqual("abc.123", renewal_info.offerIdentifier)
        self.assertEqual(1698148800000, renewal_info.signedDate)
        self.assertEqual(Environment.LOCAL_TESTING, renewal_info.environment)
        self.assertEqual("LocalTesting", renewal_info.rawEnvironment)
        self.assertEqual(1698148800000, renewal_info.recentSubscriptionStartDate)
        self.assertEqual(1698148850000, renewal_info.renewalDate)

    def test_notificaiton_decoding(self):
        signed_notification = create_signed_data_from_json('tests/resources/models/signedNotification.json')
        
        signed_data_verifier = get_default_signed_data_verifier()

        notification = signed_data_verifier.verify_and_decode_notification(signed_notification)

        self.assertEqual(NotificationTypeV2.SUBSCRIBED, notification.notificationType)
        self.assertEqual("SUBSCRIBED", notification.rawNotificationType)
        self.assertEqual(Subtype.INITIAL_BUY, notification.subtype)
        self.assertEqual("INITIAL_BUY", notification.rawSubtype)
        self.assertEqual("002e14d5-51f5-4503-b5a8-c3a1af68eb20", notification.notificationUUID)
        self.assertEqual("2.0", notification.version)
        self.assertEqual(1698148900000, notification.signedDate)
        self.assertIsNotNone(notification.data)
        self.assertIsNone(notification.summary)
        self.assertEqual(Environment.LOCAL_TESTING, notification.data.environment)
        self.assertEqual("LocalTesting", notification.data.rawEnvironment)
        self.assertEqual(41234, notification.data.appAppleId)
        self.assertEqual("com.example", notification.data.bundleId)
        self.assertEqual("1.2.3", notification.data.bundleVersion)
        self.assertEqual("signed_transaction_info_value", notification.data.signedTransactionInfo)
        self.assertEqual("signed_renewal_info_value", notification.data.signedRenewalInfo);
        self.assertEqual(Status.ACTIVE, notification.data.status)
        self.assertEqual(1, notification.data.rawStatus)
        
    def test_summary_notification_decoding(self):
        signed_summary_notification = create_signed_data_from_json('tests/resources/models/signedSummaryNotification.json')
        
        signed_data_verifier = get_default_signed_data_verifier()

        notification = signed_data_verifier.verify_and_decode_notification(signed_summary_notification)
        
        self.assertEqual(NotificationTypeV2.RENEWAL_EXTENSION, notification.notificationType)
        self.assertEqual("RENEWAL_EXTENSION", notification.rawNotificationType)
        self.assertEqual(Subtype.SUMMARY, notification.subtype)
        self.assertEqual("SUMMARY", notification.rawSubtype)
        self.assertEqual("002e14d5-51f5-4503-b5a8-c3a1af68eb20", notification.notificationUUID)
        self.assertEqual("2.0", notification.version)
        self.assertEqual(1698148900000, notification.signedDate)
        self.assertIsNone(notification.data)
        self.assertIsNotNone(notification.summary)
        self.assertEqual(Environment.LOCAL_TESTING, notification.summary.environment)
        self.assertEqual("LocalTesting", notification.summary.rawEnvironment)
        self.assertEqual(41234, notification.summary.appAppleId)
        self.assertEqual("com.example", notification.summary.bundleId)
        self.assertEqual("com.example.product", notification.summary.productId)
        self.assertEqual("efb27071-45a4-4aca-9854-2a1e9146f265", notification.summary.requestIdentifier)
        self.assertEqual(["CAN", "USA", "MEX"], notification.summary.storefrontCountryCodes)
        self.assertEqual(5, notification.summary.succeededCount)
        self.assertEqual(2, notification.summary.failedCount)


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

import unittest
from appstoreserverlibrary.promotional_offer import PromotionalOfferSignatureCreator

from tests.util import read_data_from_binary_file
from uuid import UUID


class PromotionalOfferSignatureCreatorTest(unittest.TestCase):
    def test_signature_creator(self):
        signing_key = read_data_from_binary_file('tests/resources/certs/testSigningKey.p8')
        signature_creator = PromotionalOfferSignatureCreator(signing_key, 'keyId', 'bundleId')
        signature = signature_creator.create_signature("productId", "offerId", "applicationUsername", UUID("20fba8a0-2b80-4a7d-a17f-85c1854727f8"), 1698148900000)
        self.assertIsNotNone(signature)


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

from .LibraryUtility import AttrsRawValueAware
from .SendAttemptResult import SendAttemptResult

@define
class SendAttemptItem(AttrsRawValueAware):
    """
    The success or error information and the date the App Store server records when it attempts to send a server notification to your server.
    
    https://developer.apple.com/documentation/appstoreserverapi/sendattemptitem
    """

    attemptDate: Optional[int] = attr.ib(default=None)
    """
    The date the App Store server attempts to send a notification.
    
    https://developer.apple.com/documentation/appstoreserverapi/attemptdate
    """

    sendAttemptResult: Optional[SendAttemptResult] = SendAttemptResult.create_main_attr('rawSendAttemptResult')
    """
    The success or error information the App Store server records when it attempts to send an App Store server notification to your server.
    
    https://developer.apple.com/documentation/appstoreserverapi/sendattemptresult
    """
        
    rawSendAttemptResult: Optional[str] = SendAttemptResult.create_raw_attr('sendAttemptResult')
    """
    See sendAttemptResult
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from attr import define
from typing import List, Optional
import attr

from .Environment import Environment
from .LibraryUtility import AttrsRawValueAware

@define
class Summary(AttrsRawValueAware):
    """
    The payload data for a subscription-renewal-date extension notification.
    
    https://developer.apple.com/documentation/appstoreservernotifications/summary
    """

    environment: Optional[Environment] = Environment.create_main_attr('rawEnvironment')
    """
    The server environment that the notification applies to, either sandbox or production.
    
    https://developer.apple.com/documentation/appstoreservernotifications/environment
    """
           
    rawEnvironment: Optional[str] = Environment.create_raw_attr('environment')
    """
    See environment
    """

    appAppleId: Optional[int] = attr.ib(default=None)
    """
    The unique identifier of an app in the App Store.
    
    https://developer.apple.com/documentation/appstoreservernotifications/appappleid
    """

    bundleId: Optional[str] = attr.ib(default=None)
    """
    The bundle identifier of an app.
    
    https://developer.apple.com/documentation/appstoreserverapi/bundleid
    """

    productId: Optional[str] = attr.ib(default=None)
    """
    The unique identifier for the product, that you create in App Store Connect.
    
    https://developer.apple.com/documentation/appstoreserverapi/productid
    """    

    requestIdentifier: Optional[str] = attr.ib(default=None)
    """
    A string that contains a unique identifier you provide to track each subscription-renewal-date extension request.
    
    https://developer.apple.com/documentation/appstoreserverapi/requestidentifier
    """

    storefrontCountryCodes: Optional[List[str]] = attr.ib(default=None)
    """
    A list of storefront country codes you provide to limit the storefronts for a subscription-renewal-date extension.
    
    https://developer.apple.com/documentation/appstoreserverapi/storefrontcountrycodes
    """

    succeededCount: Optional[int] = attr.ib(default=None)
    """
    The count of subscriptions that successfully receive a subscription-renewal-date extension.
    
    https://developer.apple.com/documentation/appstoreserverapi/succeededcount
    """

    failedCount: Optional[int] = attr.ib(default=None)
    """
    The count of subscriptions that fail to receive a subscription-renewal-date extension.
    
    https://developer.apple.com/documentation/appstoreserverapi/failedcount
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from attr import define
from typing import List, Optional
import attr

from .LibraryUtility import AttrsRawValueAware
from .OrderLookupStatus import OrderLookupStatus

@define
class OrderLookupResponse(AttrsRawValueAware):
    """
    A response that includes the order lookup status and an array of signed transactions for the in-app purchases in the order.
    
    https://developer.apple.com/documentation/appstoreserverapi/orderlookupresponse
    """

    status: Optional[OrderLookupStatus] = OrderLookupStatus.create_main_attr('rawStatus')
    """
    The status that indicates whether the order ID is valid.
    
    https://developer.apple.com/documentation/appstoreserverapi/orderlookupstatus
    """

    rawStatus: Optional[int] = OrderLookupStatus.create_raw_attr('status')
    """
    See status
    """

    signedTransactions: Optional[List[str]] = attr.ib(default=None)
    """
    An array of in-app purchase transactions that are part of order, signed by Apple, in JSON Web Signature format.
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class OrderLookupStatus(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    A value that indicates whether the order ID in the request is valid for your app.
    
    https://developer.apple.com/documentation/appstoreserverapi/orderlookupstatus
    """
    VALID = 0
    INVALID = 1



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

@define
class MassExtendRenewalDateStatusResponse: 
    """
    A response that indicates the current status of a request to extend the subscription renewal date to all eligible subscribers.
    
    https://developer.apple.com/documentation/appstoreserverapi/massextendrenewaldatestatusresponse
    """

    requestIdentifier: Optional[str] = attr.ib(default=None)
    """
    A string that contains a unique identifier you provide to track each subscription-renewal-date extension request.
    
    https://developer.apple.com/documentation/appstoreserverapi/requestidentifier
    """

    complete: Optional[bool] = attr.ib(default=None)
    """
    A Boolean value that indicates whether the App Store completed the request to extend a subscription renewal date to active subscribers.
    
    https://developer.apple.com/documentation/appstoreserverapi/complete
    """

    completeDate: Optional[int] = attr.ib(default=None)
    """
    The UNIX time, in milliseconds, that the App Store completes a request to extend a subscription renewal date for eligible subscribers.
    
    https://developer.apple.com/documentation/appstoreserverapi/completedate
    """

    succeededCount: Optional[int] = attr.ib(default=None)
    """
    The count of subscriptions that successfully receive a subscription-renewal-date extension.
    
    https://developer.apple.com/documentation/appstoreserverapi/succeededcount
    """

    failedCount: Optional[int] = attr.ib(default=None)
    """
    The count of subscriptions that fail to receive a subscription-renewal-date extension.
    
    https://developer.apple.com/documentation/appstoreserverapi/failedcount
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class AutoRenewStatus(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The renewal status for an auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/autorenewstatus
    """
    OFF = 0
    ON = 1



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional, List

from attr import define
import attr

from .SendAttemptItem import SendAttemptItem

@define
class NotificationHistoryResponseItem: 
    """
    The App Store server notification history record, including the signed notification payload and the result of the server's first send attempt.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationhistoryresponseitem
    """

    signedPayload: Optional[str] = attr.ib(default=None)
    """
    A cryptographically signed payload, in JSON Web Signature (JWS) format, containing the response body for a version 2 notification.
    
    https://developer.apple.com/documentation/appstoreservernotifications/signedpayload
    """

    sendAttempts: Optional[List[SendAttemptItem]] = attr.ib(default=None)
    """
    An array of information the App Store server records for its attempts to send a notification to your server. The maximum number of entries in the array is six.
    
    https://developer.apple.com/documentation/appstoreserverapi/sendattemptitem
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

from .LibraryUtility import AttrsRawValueAware
from .Status import Status

@define
class LastTransactionsItem(AttrsRawValueAware):
    """
    The most recent App Store-signed transaction information and App Store-signed renewal information for an auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/lasttransactionsitem
    """

    status: Optional[Status] = Status.create_main_attr('rawStatus')
    """
    The status of the auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/status
    """

    rawStatus: Optional[int] = Status.create_raw_attr('status')
    """
    See status
    """

    originalTransactionId: Optional[str] = attr.ib(default=None)
    """
    The original transaction identifier of a purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/originaltransactionid
    """

    signedTransactionInfo: Optional[str] = attr.ib(default=None)
    """
    Transaction information signed by the App Store, in JSON Web Signature (JWS) format.
    
    https://developer.apple.com/documentation/appstoreserverapi/jwstransaction
    """

    signedRenewalInfo: Optional[str] = attr.ib(default=None)
    """
    Subscription renewal information, signed by the App Store, in JSON Web Signature (JWS) format.
    
    https://developer.apple.com/documentation/appstoreserverapi/jwsrenewalinfo
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

import unittest
from appstoreserverlibrary.models.AutoRenewStatus import AutoRenewStatus
from appstoreserverlibrary.models.Environment import Environment
from appstoreserverlibrary.models.InAppOwnershipType import InAppOwnershipType
from appstoreserverlibrary.models.OfferType import OfferType
from appstoreserverlibrary.models.TransactionReason import TransactionReason
from appstoreserverlibrary.models.Type import Type
from appstoreserverlibrary.receipt_utility import ReceiptUtility
from appstoreserverlibrary.signed_data_verifier import VerificationException

from tests.util import get_signed_data_verifier, read_data_from_file

XCODE_BUNDLE_ID = "com.example.naturelab.backyardbirds.example"

class ReceiptUtilityTest(unittest.TestCase):
    def test_xcode_signed_app_transaction(self):
        verifier = get_signed_data_verifier(Environment.XCODE, XCODE_BUNDLE_ID)
        encoded_app_transaction = read_data_from_file("tests/resources/xcode/xcode-signed-app-transaction")

        app_transaction = verifier.verify_and_decode_app_transaction(encoded_app_transaction)

        self.assertIsNotNone(app_transaction)
        self.assertIsNone(app_transaction.appAppleId)
        self.assertEqual(XCODE_BUNDLE_ID, app_transaction.bundleId)
        self.assertEqual("1", app_transaction.applicationVersion)
        self.assertIsNone(app_transaction.versionExternalIdentifier)
        self.assertEqual(-62135769600000, app_transaction.originalPurchaseDate)
        self.assertEqual("1", app_transaction.originalApplicationVersion)
        self.assertEqual("cYUsXc53EbYc0pOeXG5d6/31LGHeVGf84sqSN0OrJi5u/j2H89WWKgS8N0hMsMlf", app_transaction.deviceVerification)
        self.assertEqual("48c8b92d-ce0d-4229-bedf-e61b4f9cfc92", app_transaction.deviceVerificationNonce)
        self.assertIsNone(app_transaction.preorderDate)
        self.assertEqual(Environment.XCODE, app_transaction.receiptType)
        self.assertEqual("Xcode", app_transaction.rawReceiptType)

    def test_xcode_signed_transaction(self):
        verifier = get_signed_data_verifier(Environment.XCODE, XCODE_BUNDLE_ID)
        encoded_transaction = read_data_from_file("tests/resources/xcode/xcode-signed-transaction")

        transaction = verifier.verify_and_decode_signed_transaction(encoded_transaction)

        self.assertEqual("0", transaction.originalTransactionId)
        self.assertEqual("0", transaction.transactionId)
        self.assertEqual("0", transaction.webOrderLineItemId)
        self.assertEqual(XCODE_BUNDLE_ID, transaction.bundleId)
        self.assertEqual("pass.premium", transaction.productId)
        self.assertEqual("6F3A93AB", transaction.subscriptionGroupIdentifier)
        self.assertEqual(1697679936049, transaction.purchaseDate)
        self.assertEqual(1697679936049, transaction.originalPurchaseDate)
        self.assertEqual(1700358336049, transaction.expiresDate)
        self.assertEqual(1, transaction.quantity)
        self.assertEqual(Type.AUTO_RENEWABLE_SUBSCRIPTION, transaction.type)
        self.assertEqual("Auto-Renewable Subscription", transaction.rawType)
        self.assertIsNone(transaction.appAccountToken)
        self.assertEqual(InAppOwnershipType.PURCHASED, transaction.inAppOwnershipType)
        self.assertEqual("PURCHASED", transaction.rawInAppOwnershipType)
        self.assertEqual(1697679936056, transaction.signedDate)
        self.assertIsNone(transaction.revocationReason)
        self.assertIsNone(transaction.revocationDate)
        self.assertFalse(transaction.isUpgraded)
        self.assertEqual(OfferType.INTRODUCTORY_OFFER, transaction.offerType)
        self.assertEqual(1, transaction.rawOfferType)
        self.assertIsNone(transaction.offerIdentifier)
        self.assertEqual(Environment.XCODE, transaction.environment)
        self.assertEqual("Xcode", transaction.rawEnvironment)
        self.assertEqual("USA", transaction.storefront)
        self.assertEqual("143441", transaction.storefrontId)
        self.assertEqual(TransactionReason.PURCHASE, transaction.transactionReason)
        self.assertEqual("PURCHASE", transaction.rawTransactionReason)

    def test_xcode_signed_renewal_info(self):
        verifier = get_signed_data_verifier(Environment.XCODE, XCODE_BUNDLE_ID)
        encoded_renewal_info = read_data_from_file("tests/resources/xcode/xcode-signed-renewal-info")

        renewal_info = verifier.verify_and_decode_renewal_info(encoded_renewal_info)

        self.assertIsNone(renewal_info.expirationIntent)
        self.assertEqual("0", renewal_info.originalTransactionId)
        self.assertEqual("pass.premium", renewal_info.autoRenewProductId)
        self.assertEqual("pass.premium", renewal_info.productId)
        self.assertEqual(AutoRenewStatus.ON, renewal_info.autoRenewStatus)
        self.assertEqual(1, renewal_info.rawAutoRenewStatus)
        self.assertIsNone(renewal_info.isInBillingRetryPeriod)
        self.assertIsNone(renewal_info.priceIncreaseStatus)
        self.assertIsNone(renewal_info.gracePeriodExpiresDate)
        self.assertIsNone(renewal_info.offerType)
        self.assertIsNone(renewal_info.offerIdentifier)
        self.assertEqual(1697679936711, renewal_info.signedDate)
        self.assertEqual(Environment.XCODE, renewal_info.environment)
        self.assertEqual("Xcode", renewal_info.rawEnvironment)
        self.assertEqual(1697679936049, renewal_info.recentSubscriptionStartDate)
        self.assertEqual(1700358336049, renewal_info.renewalDate)

    def test_xcode_signed_app_transaction_with_production_environment(self):
        verifier = get_signed_data_verifier(Environment.PRODUCTION, XCODE_BUNDLE_ID)
        encoded_app_transaction = read_data_from_file("tests/resources/xcode/xcode-signed-app-transaction")
        try:
            verifier.verify_and_decode_app_transaction(encoded_app_transaction)
        except VerificationException:
            return
        self.assertFalse(True)




# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class SendAttemptResult(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The success or error information the App Store server records when it attempts to send an App Store server notification to your server.
    
    https://developer.apple.com/documentation/appstoreserverapi/sendattemptresult
    """
    SUCCESS = "SUCCESS"
    TIMED_OUT = "TIMED_OUT"
    TLS_ISSUE = "TLS_ISSUE"
    CIRCULAR_REDIRECT = "CIRCULAR_REDIRECT"
    NO_RESPONSE = "NO_RESPONSE"
    SOCKET_ISSUE = "SOCKET_ISSUE"
    UNSUPPORTED_CHARSET = "UNSUPPORTED_CHARSET"
    INVALID_RESPONSE = "INVALID_RESPONSE"
    PREMATURE_CLOSE = "PREMATURE_CLOSE"
    UNSUCCESSFUL_HTTP_RESPONSE_CODE = "UNSUCCESSFUL_HTTP_RESPONSE_CODE"
    OTHER = "OTHER"



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from typing import Optional
from attr import define
import attr

from .NotificationTypeV2 import NotificationTypeV2
from .Subtype import Subtype

@define
class NotificationHistoryRequest: 
    """
    The request body for notification history.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationhistoryrequest
    """

    startDate: Optional[int] = attr.ib(default=None)
    """
    The start date of the timespan for the requested App Store Server Notification history records. The startDate needs to precede the endDate. Choose a startDate that's within the past 180 days from the current date.
    
    https://developer.apple.com/documentation/appstoreserverapi/startdate
    """

    endDate: Optional[int] = attr.ib(default=None)
    """
    The end date of the timespan for the requested App Store Server Notification history records. Choose an endDate that's later than the startDate. If you choose an endDate in the future, the endpoint automatically uses the current date as the endDate.
    
    https://developer.apple.com/documentation/appstoreserverapi/enddate
    """

    notificationType: Optional[NotificationTypeV2] = attr.ib(default=None)
    """
    A notification type. Provide this field to limit the notification history records to those with this one notification type. For a list of notifications types, see notificationType.
    Include either the transactionId or the notificationType in your query, but not both.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationtype
    """

    notificationSubtype: Optional[Subtype] = attr.ib(default=None)
    """
    A notification subtype. Provide this field to limit the notification history records to those with this one notification subtype. For a list of subtypes, see subtype. If you specify a notificationSubtype, you need to also specify its related notificationType.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationsubtype
    """

    transactionId: Optional[str] = attr.ib(default=None)
    """
    The transaction identifier, which may be an original transaction identifier, of any transaction belonging to the customer. Provide this field to limit the notification history request to this one customer.
    Include either the transactionId or the notificationType in your query, but not both.
    
    https://developer.apple.com/documentation/appstoreserverapi/transactionid
    """

    onlyFailures: Optional[bool] = attr.ib(default=None)
    """
    A Boolean value you set to true to request only the notifications that haven’t reached your server successfully. The response also includes notifications that the App Store server is currently retrying to send to your server.
    
    https://developer.apple.com/documentation/appstoreserverapi/onlyfailures
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional, List

from attr import define
import attr

from .LastTransactionsItem import LastTransactionsItem

@define
class SubscriptionGroupIdentifierItem: 
    """
    Information for auto-renewable subscriptions, including signed transaction information and signed renewal information, for one subscription group.
    
    https://developer.apple.com/documentation/appstoreserverapi/subscriptiongroupidentifieritem
    """

    subscriptionGroupIdentifier: Optional[str] = attr.ib(default=None)
    """
    The identifier of the subscription group that the subscription belongs to.
    
    https://developer.apple.com/documentation/appstoreserverapi/subscriptiongroupidentifier
    """    

    lastTransactions: Optional[List[LastTransactionsItem]] = attr.ib(default=None)
    """
    An array of the most recent App Store-signed transaction information and App Store-signed renewal information for all auto-renewable subscriptions in the subscription group.
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class OfferType(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The type of subscription offer.
    
    https://developer.apple.com/documentation/appstoreserverapi/offertype
    """
    INTRODUCTORY_OFFER = 1
    PROMOTIONAL_OFFER = 2
    SUBSCRIPTION_OFFER_CODE = 3



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

from .Environment import Environment
from .Status import Status
from .LibraryUtility import AttrsRawValueAware

@define
class Data(AttrsRawValueAware):
    """
    The app metadata and the signed renewal and transaction information.
    
    https://developer.apple.com/documentation/appstoreservernotifications/data
    """
    environment: Optional[Environment] = Environment.create_main_attr('rawEnvironment')
    """
    The server environment that the notification applies to, either sandbox or production.
    
    https://developer.apple.com/documentation/appstoreservernotifications/environment
    """
    rawEnvironment: Optional[str] = Environment.create_raw_attr('environment')
    """
    See environment
    """

    appAppleId: Optional[int] = attr.ib(default=None)
    """
    The unique identifier of an app in the App Store.
    
    https://developer.apple.com/documentation/appstoreservernotifications/appappleid
    """

    bundleId: Optional[str] = attr.ib(default=None)
    """
    The bundle identifier of an app.
    
    https://developer.apple.com/documentation/appstoreserverapi/bundleid
    """

    bundleVersion: Optional[str] = attr.ib(default=None)
    """
    The version of the build that identifies an iteration of the bundle.
    
    https://developer.apple.com/documentation/appstoreservernotifications/bundleversion
    """

    signedTransactionInfo: Optional[str] = attr.ib(default=None)
    """
    Transaction information signed by the App Store, in JSON Web Signature (JWS) format.
    
    https://developer.apple.com/documentation/appstoreserverapi/jwstransaction
    """

    signedRenewalInfo: Optional[str] = attr.ib(default=None)
    """
    Subscription renewal information, signed by the App Store, in JSON Web Signature (JWS) format.
    
    https://developer.apple.com/documentation/appstoreserverapi/jwsrenewalinfo
    """
        
    status: Optional[Status] = Status.create_main_attr('rawStatus')
    """
    The status of an auto-renewable subscription as of the signedDate in the responseBodyV2DecodedPayload.
    
    https://developer.apple.com/documentation/appstoreservernotifications/status
    """

    rawStatus: Optional[int] = Status.create_raw_attr('status')
    """
    See status
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

from .AccountTenure import AccountTenure
from .ConsumptionStatus import ConsumptionStatus
from .DeliveryStatus import DeliveryStatus
from .LibraryUtility import AttrsRawValueAware
from .LifetimeDollarsPurchased import LifetimeDollarsPurchased
from .LifetimeDollarsRefunded import LifetimeDollarsRefunded
from .Platform import Platform
from .PlayTime import PlayTime
from .UserStatus import UserStatus

@define
class ConsumptionRequest(AttrsRawValueAware):
    """
    The request body containing consumption information.
    
    https://developer.apple.com/documentation/appstoreserverapi/consumptionrequest
    """

    customerConsented: Optional[bool] = attr.ib(default=None)
    """
    A Boolean value that indicates whether the customer consented to provide consumption data to the App Store.
    
    https://developer.apple.com/documentation/appstoreserverapi/customerconsented
    """

    consumptionStatus: Optional[ConsumptionStatus] = ConsumptionStatus.create_main_attr('rawConsumptionStatus')
    """
    A value that indicates the extent to which the customer consumed the in-app purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/consumptionstatus
    """

    rawConsumptionStatus: Optional[int] = ConsumptionStatus.create_raw_attr('consumptionStatus')
    """
    See consumptionStatus
    """

    platform: Optional[Platform] = Platform.create_main_attr('rawPlatform')
    """
    A value that indicates the platform on which the customer consumed the in-app purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/platform
    """

    rawPlatform: Optional[int] = Platform.create_raw_attr('platform')
    """
    See platform
    """

    sampleContentProvided: Optional[bool] = attr.ib(default=None)
    """
    A Boolean value that indicates whether you provided, prior to its purchase, a free sample or trial of the content, or information about its functionality.
    
    https://developer.apple.com/documentation/appstoreserverapi/samplecontentprovided
    """

    deliveryStatus: Optional[DeliveryStatus] = DeliveryStatus.create_main_attr('rawDeliveryStatus')
    """
    A value that indicates whether the app successfully delivered an in-app purchase that works properly.
    
    https://developer.apple.com/documentation/appstoreserverapi/deliverystatus
    """

    rawDeliveryStatus: Optional[int] = DeliveryStatus.create_raw_attr('deliveryStatus')
    """
    See deliveryStatus
    """

    appAccountToken: Optional[str] = attr.ib(default=None)
    """
    The UUID that an app optionally generates to map a customer's in-app purchase with its resulting App Store transaction.
    
    https://developer.apple.com/documentation/appstoreserverapi/appaccounttoken
    """

    accountTenure: Optional[AccountTenure] = AccountTenure.create_main_attr('rawAccountTenure')
    """
    The age of the customer's account.
    
    https://developer.apple.com/documentation/appstoreserverapi/accounttenure
    """

    rawAccountTenure: Optional[int] = AccountTenure.create_raw_attr('accountTenure')
    """
    See accountTenure
    """

    playTime: Optional[PlayTime] = PlayTime.create_main_attr('rawPlayTime')
    """
    A value that indicates the amount of time that the customer used the app.
    
    https://developer.apple.com/documentation/appstoreserverapi/consumptionrequest
    """

    rawPlayTime: Optional[int] = PlayTime.create_raw_attr('playTime')
    """
    See playTime
    """

    lifetimeDollarsRefunded: Optional[LifetimeDollarsRefunded] = LifetimeDollarsRefunded.create_main_attr('rawLifetimeDollarsRefunded')
    """
    A value that indicates the total amount, in USD, of refunds the customer has received, in your app, across all platforms.
    
    https://developer.apple.com/documentation/appstoreserverapi/lifetimedollarsrefunded
    """

    rawLifetimeDollarsRefunded: Optional[int] = LifetimeDollarsRefunded.create_raw_attr('lifetimeDollarsRefunded')
    """
    See lifetimeDollarsRefunded
    """

    lifetimeDollarsPurchased: Optional[LifetimeDollarsPurchased] = LifetimeDollarsPurchased.create_main_attr('rawLifetimeDollarsPurchased')
    """
    A value that indicates the total amount, in USD, of in-app purchases the customer has made in your app, across all platforms.
    
    https://developer.apple.com/documentation/appstoreserverapi/lifetimedollarspurchased
    """

    rawLifetimeDollarsPurchased: Optional[int] = LifetimeDollarsPurchased.create_raw_attr('lifetimeDollarsPurchased')
    """
    See lifetimeDollarsPurchased
    """

    userStatus: Optional[UserStatus] =  UserStatus.create_main_attr('rawUserStatus')
    """
    The status of the customer's account.
    
    https://developer.apple.com/documentation/appstoreserverapi/userstatus
    """

    rawUserStatus: Optional[int] = UserStatus.create_raw_attr('userStatus')
    """
    See userStatus
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class Platform(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The platform on which the customer consumed the in-app purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/platform
    """
    UNDECLARED = 0
    APPLE = 1
    NON_APPLE = 2



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class Environment(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The server environment, either sandbox or production.
    
    https://developer.apple.com/documentation/appstoreserverapi/environment
    """
    SANDBOX = "Sandbox"
    PRODUCTION = "Production"
    XCODE = "Xcode"
    LOCAL_TESTING = "LocalTesting"



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class ExpirationIntent(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The reason an auto-renewable subscription expired.
    
    https://developer.apple.com/documentation/appstoreserverapi/expirationintent
    """
    CUSTOMER_CANCELLED = 1
    BILLING_ERROR = 2
    CUSTOMER_DID_NOT_CONSENT_TO_PRICE_INCREASE = 3
    PRODUCT_NOT_AVAILABLE = 4
    OTHER = 5



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum
from typing import List, Optional
import attr


from .InAppOwnershipType import InAppOwnershipType

class ProductType(str, Enum):
    AUTO_RENEWABLE = "AUTO_RENEWABLE"
    NON_RENEWABLE = "NON_RENEWABLE"
    CONSUMABLE = "CONSUMABLE"
    NON_CONSUMABLE = "NON_CONSUMABLE"

class Order(str, Enum):
    ASCENDING = "ASCENDING"
    DESCENDING = "DESCENDING"

@attr.define
class TransactionHistoryRequest:
    startDate: Optional[int] = attr.ib(default=None)
    """
    An optional start date of the timespan for the transaction history records you're requesting. The startDate must precede the endDate if you specify both dates. To be included in results, the transaction's purchaseDate must be equal to or greater than the startDate.
    
    https://developer.apple.com/documentation/appstoreserverapi/startdate
    """
    
    endDate: Optional[int] = attr.ib(default=None)
    """
    An optional end date of the timespan for the transaction history records you're requesting. Choose an endDate that's later than the startDate if you specify both dates. Using an endDate in the future is valid. To be included in results, the transaction's purchaseDate must be less than the endDate.
    
    https://developer.apple.com/documentation/appstoreserverapi/enddate
    """

    productIds: Optional[List[str]] = attr.ib(default=None)
    """
    An optional filter that indicates the product identifier to include in the transaction history. Your query may specify more than one productID.
    
    https://developer.apple.com/documentation/appstoreserverapi/productid
    """    

    productTypes: Optional[List[ProductType]] = attr.ib(default=None)
    """
    An optional filter that indicates the product type to include in the transaction history. Your query may specify more than one productType.
    """
    
    sort: Optional[Order] = attr.ib(default=None)
    """
    An optional sort order for the transaction history records. The response sorts the transaction records by their recently modified date. The default value is ASCENDING, so you receive the oldest records first.
    """

    subscriptionGroupIdentifiers: Optional[List[str]] = attr.ib(default=None)
    """
    An optional filter that indicates the subscription group identifier to include in the transaction history. Your query may specify more than one subscriptionGroupIdentifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/subscriptiongroupidentifier
    """

    inAppOwnershipType: Optional[InAppOwnershipType] = attr.ib(default=None)
    """
    An optional filter that limits the transaction history by the in-app ownership type.
    
    https://developer.apple.com/documentation/appstoreserverapi/inappownershiptype
    """
   
    revoked: Optional[bool] = attr.ib(default=None)
    """
    An optional Boolean value that indicates whether the response includes only revoked transactions when the value is true, or contains only nonrevoked transactions when the value is false. By default, the request doesn't include this parameter.
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

@define
class ResponseBodyV2:
    """
    The response body the App Store sends in a version 2 server notification.

    https://developer.apple.com/documentation/appstoreservernotifications/responsebodyv2
    """
     
    signedPayload: Optional[str] = attr.ib(default=None)
    """
    A cryptographically signed payload, in JSON Web Signature (JWS) format, containing the response body for a version 2 notification.
    
    https://developer.apple.com/documentation/appstoreservernotifications/signedpayload
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class ConsumptionStatus(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    A value that indicates the extent to which the customer consumed the in-app purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/consumptionstatus
    """
    UNDECLARED = 0
    NOT_CONSUMED = 1
    PARTIALLY_CONSUMED = 2
    FULLY_CONSUMED = 3



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class UserStatus(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The status of a customer's account within your app.
    
    https://developer.apple.com/documentation/appstoreserverapi/userstatus
    """
    UNDECLARED = 0
    ACTIVE = 1
    SUSPENDED = 2
    TERMINATED = 3
    LIMITED_ACCESS = 4



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

import unittest
from base64 import b64decode
from appstoreserverlibrary.models.Environment import Environment
from appstoreserverlibrary.models.NotificationHistoryRequest import NotificationTypeV2

from appstoreserverlibrary.signed_data_verifier import VerificationException, VerificationStatus, SignedDataVerifier

from tests.util import get_signed_data_verifier, read_data_from_file

class PayloadVerification(unittest.TestCase):
    def test_app_store_server_notification_decoding(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.example")
        test_notification = read_data_from_file('tests/resources/mock_signed_data/testNotification')
        notification = verifier.verify_and_decode_notification(test_notification)
        self.assertEqual(notification.notificationType, NotificationTypeV2.TEST)

    def test_app_store_server_notification_decoding_production(self):
        verifier = get_signed_data_verifier(Environment.PRODUCTION, "com.example")
        test_notification = read_data_from_file('tests/resources/mock_signed_data/testNotification')
        with self.assertRaises(VerificationException) as context:
            verifier.verify_and_decode_notification(test_notification)
        self.assertEqual(context.exception.status, VerificationStatus.INVALID_ENVIRONMENT)

    def test_missing_x5c_header(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.example")
        missing_x5c_header_claim = read_data_from_file('tests/resources/mock_signed_data/missingX5CHeaderClaim')
        with self.assertRaises(VerificationException) as context:
            verifier.verify_and_decode_notification(missing_x5c_header_claim)
        self.assertEqual(context.exception.status, VerificationStatus.VERIFICATION_FAILURE)

    def test_wrong_bundle_id_for_server_notification(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.examplex")
        wrong_bundle = read_data_from_file('tests/resources/mock_signed_data/wrongBundleId')
        with self.assertRaises(VerificationException) as context:
            verifier.verify_and_decode_notification(wrong_bundle)
        self.assertEqual(context.exception.status, VerificationStatus.INVALID_APP_IDENTIFIER)

    def test_wrong_app_apple_id_for_server_notification(self):
        verifier = get_signed_data_verifier(Environment.PRODUCTION, "com.example", 1235)
        test_notification = read_data_from_file('tests/resources/mock_signed_data/testNotification')
        with self.assertRaises(VerificationException) as context:
            verifier.verify_and_decode_notification(test_notification)
        self.assertEqual(context.exception.status, VerificationStatus.INVALID_APP_IDENTIFIER)

    def test_renewal_info_decoding(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.example")
        renewal_info = read_data_from_file('tests/resources/mock_signed_data/renewalInfo')
        notification = verifier.verify_and_decode_renewal_info(renewal_info)
        self.assertEqual(notification.environment, Environment.SANDBOX)

    def test_transaction_info_decoding(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.example")
        transaction_info = read_data_from_file('tests/resources/mock_signed_data/transactionInfo')
        notification = verifier.verify_and_decode_signed_transaction(transaction_info)
        self.assertEqual(notification.environment, Environment.SANDBOX)

    def test_malformed_jwt_with_too_many_parts(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.example")
        with self.assertRaises(VerificationException) as context:
            verifier.verify_and_decode_notification("a.b.c.d")
        self.assertEqual(context.exception.status, VerificationStatus.VERIFICATION_FAILURE)

    def test_malformed_jwt_with_malformed_data(self):
        verifier = get_signed_data_verifier(Environment.SANDBOX, "com.example")
        with self.assertRaises(VerificationException) as context:
            verifier.verify_and_decode_notification("a.b.c")
        self.assertEqual(context.exception.status, VerificationStatus.VERIFICATION_FAILURE)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class Subtype(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    A notification subtype value that App Store Server Notifications 2 uses.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationsubtype
    """
    INITIAL_BUY = "INITIAL_BUY"
    RESUBSCRIBE = "RESUBSCRIBE"
    DOWNGRADE = "DOWNGRADE"
    UPGRADE = "UPGRADE"
    AUTO_RENEW_ENABLED = "AUTO_RENEW_ENABLED"
    AUTO_RENEW_DISABLED = "AUTO_RENEW_DISABLED"
    VOLUNTARY = "VOLUNTARY"
    BILLING_RETRY = "BILLING_RETRY"
    PRICE_INCREASE = "PRICE_INCREASE"
    GRACE_PERIOD = "GRACE_PERIOD"
    PENDING = "PENDING"
    ACCEPTED = "ACCEPTED"
    BILLING_RECOVERY = "BILLING_RECOVERY"
    PRODUCT_NOT_FOR_SALE = "PRODUCT_NOT_FOR_SALE"
    SUMMARY = "SUMMARY"
    FAILURE = "FAILURE"



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class LifetimeDollarsPurchased(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    A value that indicates the total amount, in USD, of in-app purchases the customer has made in your app, across all platforms.
    
    https://developer.apple.com/documentation/appstoreserverapi/lifetimedollarspurchased
    """
    UNDECLARED = 0
    ZERO_DOLLARS = 1
    ONE_CENT_TO_FORTY_NINE_DOLLARS_AND_NINETY_NINE_CENTS = 2
    FIFTY_DOLLARS_TO_NINETY_NINE_DOLLARS_AND_NINETY_NINE_CENTS = 3
    ONE_HUNDRED_DOLLARS_TO_FOUR_HUNDRED_NINETY_NINE_DOLLARS_AND_NINETY_NINE_CENTS = 4
    FIVE_HUNDRED_DOLLARS_TO_NINE_HUNDRED_NINETY_NINE_DOLLARS_AND_NINETY_NINE_CENTS = 5
    ONE_THOUSAND_DOLLARS_TO_ONE_THOUSAND_NINE_HUNDRED_NINETY_NINE_DOLLARS_AND_NINETY_NINE_CENTS = 6
    TWO_THOUSAND_DOLLARS_OR_GREATER = 7



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

import unittest
from appstoreserverlibrary.receipt_utility import ReceiptUtility

from tests.util import read_data_from_file

APP_RECEIPT_EXPECTED_TRANSACTION_ID = "0"
TRANSACTION_RECEIPT_EXPECTED_TRANSACTION_ID = "33993399"

class ReceiptUtilityTest(unittest.TestCase):
    def test_xcode_app_receipt_extraction_with_no_transactions(self):
        receipt = read_data_from_file("tests/resources/xcode/xcode-app-receipt-empty")

        receipt_util = ReceiptUtility()

        extracted_transaction_id = receipt_util.extract_transaction_id_from_app_receipt(receipt)

        self.assertIsNone(extracted_transaction_id)

    def test_xcode_app_receipt_extraction_with_transactions(self):
        receipt = read_data_from_file("tests/resources/xcode/xcode-app-receipt-with-transaction")

        receipt_util = ReceiptUtility()

        extracted_transaction_id = receipt_util.extract_transaction_id_from_app_receipt(receipt)

        self.assertEqual(APP_RECEIPT_EXPECTED_TRANSACTION_ID, extracted_transaction_id)

    def test_transaction_receipt_extraction(self):
        receipt = read_data_from_file("tests/resources/mock_signed_data/legacyTransaction")

        receipt_util = ReceiptUtility()

        extracted_transaction_id = receipt_util.extract_transaction_id_from_transaction_receipt(receipt)

        self.assertEqual(TRANSACTION_RECEIPT_EXPECTED_TRANSACTION_ID, extracted_transaction_id)


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class NotificationTypeV2(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    A notification type value that App Store Server Notifications V2 uses.
    
    https://developer.apple.com/documentation/appstoreserverapi/notificationtype
    """
    SUBSCRIBED = "SUBSCRIBED"
    DID_CHANGE_RENEWAL_PREF = "DID_CHANGE_RENEWAL_PREF"
    DID_CHANGE_RENEWAL_STATUS = "DID_CHANGE_RENEWAL_STATUS"
    OFFER_REDEEMED = "OFFER_REDEEMED"
    DID_RENEW = "DID_RENEW"
    EXPIRED = "EXPIRED"
    DID_FAIL_TO_RENEW = "DID_FAIL_TO_RENEW"
    GRACE_PERIOD_EXPIRED = "GRACE_PERIOD_EXPIRED"
    PRICE_INCREASE = "PRICE_INCREASE"
    REFUND = "REFUND"
    REFUND_DECLINED = "REFUND_DECLINED"
    CONSUMPTION_REQUEST = "CONSUMPTION_REQUEST"
    RENEWAL_EXTENDED = "RENEWAL_EXTENDED"
    REVOKE = "REVOKE"
    TEST = "TEST"
    RENEWAL_EXTENSION = "RENEWAL_EXTENSION"
    REFUND_REVERSED = "REFUND_REVERSED"



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

import calendar
import datetime
from enum import IntEnum
from typing import Any, Dict, List, Optional, Type, TypeVar, Union
from attr import define
import requests

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from appstoreserverlibrary.models.LibraryUtility import _get_cattrs_converter
from .models.CheckTestNotificationResponse import CheckTestNotificationResponse
from .models.ConsumptionRequest import ConsumptionRequest

from .models.Environment import Environment
from .models.ExtendRenewalDateRequest import ExtendRenewalDateRequest
from .models.ExtendRenewalDateResponse import ExtendRenewalDateResponse
from .models.HistoryResponse import HistoryResponse
from .models.MassExtendRenewalDateRequest import MassExtendRenewalDateRequest
from .models.MassExtendRenewalDateResponse import MassExtendRenewalDateResponse
from .models.MassExtendRenewalDateStatusResponse import MassExtendRenewalDateStatusResponse
from .models.NotificationHistoryRequest import NotificationHistoryRequest
from .models.NotificationHistoryResponse import NotificationHistoryResponse
from .models.OrderLookupResponse import OrderLookupResponse
from .models.RefundHistoryResponse import RefundHistoryResponse
from .models.SendTestNotificationResponse import SendTestNotificationResponse
from .models.Status import Status
from .models.StatusResponse import StatusResponse
from .models.TransactionHistoryRequest import TransactionHistoryRequest
from .models.TransactionInfoResponse import TransactionInfoResponse

T = TypeVar('T')

class APIError(IntEnum):
    GENERAL_BAD_REQUEST = 4000000
    """
    An error that indicates an invalid request.
    
    https://developer.apple.com/documentation/appstoreserverapi/generalbadrequesterror
    """

    INVALID_APP_IDENTIFIER = 4000002
    """
    An error that indicates an invalid app identifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidappidentifiererror
    """

    INVALID_REQUEST_REVISION = 4000005
    """
    An error that indicates an invalid request revision.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidrequestrevisionerror
    """

    INVALID_TRANSACTION_ID = 4000006
    """
    An error that indicates an invalid transaction identifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidtransactioniderror
    """

    INVALID_ORIGINAL_TRANSACTION_ID = 4000008
    """
    An error that indicates an invalid original transaction identifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidoriginaltransactioniderror
    """

    INVALID_EXTEND_BY_DAYS = 4000009
    """
    An error that indicates an invalid extend-by-days value.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidextendbydayserror
    """

    INVALID_EXTEND_REASON_CODE = 4000010
    """
    An error that indicates an invalid reason code.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidextendreasoncodeerror
    """

    INVALID_REQUEST_IDENTIFIER = 4000011
    """
    An error that indicates an invalid request identifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidrequestidentifiererror
    """

    START_DATE_TOO_FAR_IN_PAST = 4000012
    """
    An error that indicates that the start date is earlier than the earliest allowed date.
    
    https://developer.apple.com/documentation/appstoreserverapi/startdatetoofarinpasterror
    """

    START_DATE_AFTER_END_DATE = 4000013
    """
    An error that indicates that the end date precedes the start date, or the two dates are equal.
    
    https://developer.apple.com/documentation/appstoreserverapi/startdateafterenddateerror
    """

    INVALID_PAGINATION_TOKEN = 4000014
    """
    An error that indicates the pagination token is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidpaginationtokenerror
    """

    INVALID_START_DATE = 4000015
    """
    An error that indicates the start date is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidstartdateerror
    """

    INVALID_END_DATE = 4000016
    """
    An error that indicates the end date is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidenddateerror
    """
    
    PAGINATION_TOKEN_EXPIRED = 4000017
    """
    An error that indicates the pagination token expired.
    
    https://developer.apple.com/documentation/appstoreserverapi/paginationtokenexpirederror
    """

    INVALID_NOTIFICATION_TYPE = 4000018
    """
    An error that indicates the notification type or subtype is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidnotificationtypeerror
    """

    MULTIPLE_FILTERS_SUPPLIED = 4000019
    """
    An error that indicates the request is invalid because it has too many constraints applied.
    
    https://developer.apple.com/documentation/appstoreserverapi/multiplefilterssuppliederror
    """

    INVALID_TEST_NOTIFICATION_TOKEN = 4000020
    """
    An error that indicates the test notification token is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidtestnotificationtokenerror
    """

    INVALID_SORT = 4000021
    """
    An error that indicates an invalid sort parameter.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidsorterror
    """

    INVALID_PRODUCT_TYPE = 4000022
    """
    An error that indicates an invalid product type parameter.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidproducttypeerror
    """

    INVALID_PRODUCT_ID = 4000023
    """
    An error that indicates the product ID parameter is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidproductiderror
    """

    INVALID_SUBSCRIPTION_GROUP_IDENTIFIER = 4000024
    """
    An error that indicates an invalid subscription group identifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidsubscriptiongroupidentifiererror
    """

    INVALID_EXCLUDE_REVOKED = 4000025
    """
    An error that indicates the query parameter exclude-revoked is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidexcluderevokederror

    .. deprecated:: 1.5
    """

    INVALID_IN_APP_OWNERSHIP_TYPE = 4000026
    """
    An error that indicates an invalid in-app ownership type parameter.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidinappownershiptypeerror
    """

    INVALID_EMPTY_STOREFRONT_COUNTRY_CODE_LIST = 4000027
    """
    An error that indicates a required storefront country code is empty.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidemptystorefrontcountrycodelisterror
    """

    INVALID_STOREFRONT_COUNTRY_CODE = 4000028
    """
    An error that indicates a storefront code is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidstorefrontcountrycodeerror
    """

    INVALID_REVOKED = 4000030
    """
    An error that indicates the revoked parameter contains an invalid value.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidrevokederror
    """

    INVALID_STATUS = 4000031
    """
    An error that indicates the status parameter is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidstatuserror
    """

    INVALID_ACCOUNT_TENURE = 4000032
    """
    An error that indicates the value of the account tenure field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidaccounttenureerror
    """

    INVALID_APP_ACCOUNT_TOKEN = 4000033
    """
    An error that indicates the value of the app account token field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidappaccounttokenerror
    """

    INVALID_CONSUMPTION_STATUS = 4000034
    """
    An error that indicates the value of the consumption status field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidconsumptionstatuserror
    """

    INVALID_CUSTOMER_CONSENTED = 4000035
    """
    An error that indicates the customer consented field is invalid or doesn’t indicate that the customer consented.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidcustomerconsentederror
    """

    INVALID_DELIVERY_STATUS = 4000036
    """
    An error that indicates the value in the delivery status field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invaliddeliverystatuserror
    """

    INVALID_LIFETIME_DOLLARS_PURCHASED = 4000037
    """
    An error that indicates the value in the lifetime dollars purchased field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidlifetimedollarspurchasederror
    """

    INVALID_LIFETIME_DOLLARS_REFUNDED = 4000038
    """
    An error that indicates the value in the lifetime dollars refunded field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidlifetimedollarsrefundederror
    """

    INVALID_PLATFORM = 4000039
    """
    An error that indicates the value in the platform field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidplatformerror
    """

    INVALID_PLAY_TIME = 4000040
    """
    An error that indicates the value in the playtime field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidplaytimeerror
    """

    INVALID_SAMPLE_CONTENT_PROVIDED = 4000041
    """
    An error that indicates the value in the sample content provided field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidsamplecontentprovidederror
    """

    INVALID_USER_STATUS = 4000042
    """
    An error that indicates the value in the user status field is invalid.
    
    https://developer.apple.com/documentation/appstoreserverapi/invaliduserstatuserror
    """

    INVALID_TRANSACTION_NOT_CONSUMABLE = 4000043
    """
    An error that indicates the transaction identifier doesn’t represent a consumable in-app purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/invalidtransactionnotconsumableerror
    """

    SUBSCRIPTION_EXTENSION_INELIGIBLE = 4030004
    """
    An error that indicates the subscription doesn't qualify for a renewal-date extension due to its subscription state.
    
    https://developer.apple.com/documentation/appstoreserverapi/subscriptionextensionineligibleerror
    """

    SUBSCRIPTION_MAX_EXTENSION = 4030005
    """
    An error that indicates the subscription doesn’t qualify for a renewal-date extension because it has already received the maximum extensions.
    
    https://developer.apple.com/documentation/appstoreserverapi/subscriptionmaxextensionerror
    """

    FAMILY_SHARED_SUBSCRIPTION_EXTENSION_INELIGIBLE = 4030007
    """
    An error that indicates a subscription isn't directly eligible for a renewal date extension because the user obtained it through Family Sharing.
    
    https://developer.apple.com/documentation/appstoreserverapi/familysharedsubscriptionextensionineligibleerror
    """

    ACCOUNT_NOT_FOUND = 4040001
    """
    An error that indicates the App Store account wasn’t found.
    
    https://developer.apple.com/documentation/appstoreserverapi/accountnotfounderror
    """

    ACCOUNT_NOT_FOUND_RETRYABLE = 4040002
    """
    An error response that indicates the App Store account wasn’t found, but you can try again.
    
    https://developer.apple.com/documentation/appstoreserverapi/accountnotfoundretryableerror
    """

    APP_NOT_FOUND = 4040003
    """
    An error that indicates the app wasn’t found.
    
    https://developer.apple.com/documentation/appstoreserverapi/appnotfounderror
    """

    APP_NOT_FOUND_RETRYABLE = 4040004
    """
    An error response that indicates the app wasn’t found, but you can try again.
    
    https://developer.apple.com/documentation/appstoreserverapi/appnotfoundretryableerror
    """

    ORIGINAL_TRANSACTION_ID_NOT_FOUND = 4040005
    """
    An error that indicates an original transaction identifier wasn't found.
    
    https://developer.apple.com/documentation/appstoreserverapi/originaltransactionidnotfounderror
    """

    ORIGINAL_TRANSACTION_ID_NOT_FOUND_RETRYABLE = 4040006
    """
    An error response that indicates the original transaction identifier wasn’t found, but you can try again.
    
    https://developer.apple.com/documentation/appstoreserverapi/originaltransactionidnotfoundretryableerror
    """

    SERVER_NOTIFICATION_URL_NOT_FOUND = 4040007
    """
    An error that indicates that the App Store server couldn’t find a notifications URL for your app in this environment.
    
    https://developer.apple.com/documentation/appstoreserverapi/servernotificationurlnotfounderror
    """

    TEST_NOTIFICATION_NOT_FOUND = 4040008
    """
    An error that indicates that the test notification token is expired or the test notification status isn’t available.
    
    https://developer.apple.com/documentation/appstoreserverapi/testnotificationnotfounderror
    """

    STATUS_REQUEST_NOT_FOUND = 4040009
    """
    An error that indicates the server didn't find a subscription-renewal-date extension request for the request identifier and product identifier you provided.
    
    https://developer.apple.com/documentation/appstoreserverapi/statusrequestnotfounderror
    """

    TRANSACTION_ID_NOT_FOUND = 4040010
    """
    An error that indicates a transaction identifier wasn't found.
    
    https://developer.apple.com/documentation/appstoreserverapi/transactionidnotfounderror
    """

    RATE_LIMIT_EXCEEDED = 4290000
    """
    An error that indicates that the request exceeded the rate limit.
    
    https://developer.apple.com/documentation/appstoreserverapi/ratelimitexceedederror
    """

    GENERAL_INTERNAL = 5000000
    """
    An error that indicates a general internal error.
    
    https://developer.apple.com/documentation/appstoreserverapi/generalinternalerror
    """

    GENERAL_INTERNAL_RETRYABLE = 5000001
    """
    An error response that indicates an unknown error occurred, but you can try again.
    
    https://developer.apple.com/documentation/appstoreserverapi/generalinternalretryableerror
    """


@define
class APIException(Exception):
    http_status_code: int
    api_error: Optional[APIError]
    raw_api_error: Optional[int]
    error_message: Optional[str]

    def __init__(self, http_status_code: int, raw_api_error: Optional[int] = None, error_message: Optional[str] = None):
        self.http_status_code = http_status_code
        self.raw_api_error = raw_api_error
        self.api_error = None
        self.error_message = error_message
        try:
            if raw_api_error is not None:
                self.api_error = APIError(raw_api_error)
        except ValueError:
            pass

class AppStoreServerAPIClient:
    def __init__(self, signing_key: bytes, key_id: str, issuer_id: str, bundle_id: str, environment: Environment):
        if environment == Environment.PRODUCTION:
            self._base_url = "https://api.storekit.itunes.apple.com"
        else:
            self._base_url = "https://api.storekit-sandbox.itunes.apple.com"
        self._signing_key = serialization.load_pem_private_key(signing_key, password=None, backend=default_backend())
        self._key_id = key_id
        self._issuer_id = issuer_id
        self._bundle_id = bundle_id

    def _generate_token(self) -> str:
        future_time = datetime.datetime.utcnow() + datetime.timedelta(minutes=5)
        return jwt.encode(
            {
                "bid": self._bundle_id,
                "iss": self._issuer_id,
                "aud": "appstoreconnect-v1",
                "exp": calendar.timegm(future_time.timetuple()),
            },
            self._signing_key,
            algorithm="ES256",
            headers={"kid": self._key_id},
        )

    def _make_request(self, path: str, method: str, queryParameters: Dict[str, Union[str, List[str]]], body, destination_class: Type[T]) -> T:
        url = self._base_url + path
        c = _get_cattrs_converter(type(body)) if body != None else None
        json = c.unstructure(body) if body != None else None
        headers = {
            'User-Agent': "app-store-server-library/python/1.0.0",
            'Authorization': 'Bearer ' + self._generate_token(),
            'Accept': 'application/json'
        }
        
        response = self._execute_request(method, url, queryParameters, headers, json)
        if response.status_code >= 200 and response.status_code < 300:
            if destination_class == None:
                return
            c = _get_cattrs_converter(destination_class)
            response_body = response.json()
            return c.structure(response_body, destination_class)
        else:
            # Best effort parsing of the response body
            if not 'content-type' in response.headers or response.headers['content-type'] != 'application/json':
                raise APIException(response.status_code)
            try:
                response_body = response.json()
                raise APIException(response.status_code, response_body['errorCode'], response_body['errorMessage'])
            except APIException as e:
                raise e
            except Exception as e:
                raise APIException(response.status_code) from e

    def _execute_request(self, method: str, url: str, params: Dict[str, Union[str, List[str]]], headers: Dict[str, str], json: Dict[str, Any]) -> requests.Response:
        return requests.request(method, url, params=params, headers=headers, json=json)

    def extend_renewal_date_for_all_active_subscribers(self, mass_extend_renewal_date_request: MassExtendRenewalDateRequest) -> MassExtendRenewalDateResponse: 
        """
        Uses a subscription's product identifier to extend the renewal date for all of its eligible active subscribers.
        https://developer.apple.com/documentation/appstoreserverapi/extend_subscription_renewal_dates_for_all_active_subscribers
        
        :param mass_extend_renewal_date_request: The request body for extending a subscription renewal date for all of its active subscribers.
        :return: A response that indicates the server successfully received the subscription-renewal-date extension request.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/subscriptions/extend/mass", "POST", {}, mass_extend_renewal_date_request, MassExtendRenewalDateResponse)

    def extend_subscription_renewal_date(self, original_transaction_id: str, extend_renewal_date_request: ExtendRenewalDateRequest) -> ExtendRenewalDateResponse:
        """
        Extends the renewal date of a customer's active subscription using the original transaction identifier.
        https://developer.apple.com/documentation/appstoreserverapi/extend_a_subscription_renewal_date
        
        :param original_transaction_id:    The original transaction identifier of the subscription receiving a renewal date extension.
        :param extend_renewal_date_request: The request body containing subscription-renewal-extension data.
        :return: A response that indicates whether an individual renewal-date extension succeeded, and related details.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/subscriptions/extend/" + original_transaction_id, "PUT", {}, extend_renewal_date_request, ExtendRenewalDateResponse)
    
    def get_all_subscription_statuses(self, transaction_id: str, status: List[Status] = None) -> StatusResponse:
        """
        Get the statuses for all of a customer's auto-renewable subscriptions in your app.
        https://developer.apple.com/documentation/appstoreserverapi/get_all_subscription_statuses
        
        :param transaction_id: The identifier of a transaction that belongs to the customer, and which may be an original transaction identifier.
        :param status: An optional filter that indicates the status of subscriptions to include in the response. Your query may specify more than one status query parameter.
        :return: A response that contains status information for all of a customer's auto-renewable subscriptions in your app.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        queryParameters: Dict[str, List[str]] = dict()
        if status != None:
            queryParameters["status"] = [s.value for s in status]
        
        return self._make_request("/inApps/v1/subscriptions/" + transaction_id, "GET", queryParameters, None, StatusResponse)
    
    def get_refund_history(self, transaction_id: str, revision: str) -> RefundHistoryResponse:
        """
        Get a paginated list of all of a customer's refunded in-app purchases for your app.
        https://developer.apple.com/documentation/appstoreserverapi/get_refund_history

        :param transaction_id: The identifier of a transaction that belongs to the customer, and which may be an original transaction identifier.
        :param revision:              A token you provide to get the next set of up to 20 transactions. All responses include a revision token. Use the revision token from the previous RefundHistoryResponse.
        :return: A response that contains status information for all of a customer's auto-renewable subscriptions in your app.
        :throws APIException: If a response was returned indicating the request could not be processed
        """

        queryParameters: Dict[str, List[str]] = dict()
        if revision != None:
            queryParameters["revision"] = [revision]
        
        return self._make_request("/inApps/v2/refund/lookup/" + transaction_id, "GET", queryParameters, None, RefundHistoryResponse)
    
    def get_status_of_subscription_renewal_date_extensions(self, request_identifier: str, product_id: str) -> MassExtendRenewalDateStatusResponse:
        """
        Checks whether a renewal date extension request completed, and provides the final count of successful or failed extensions.
        https://developer.apple.com/documentation/appstoreserverapi/get_status_of_subscription_renewal_date_extensions

        :param request_identifier: The UUID that represents your request to the Extend Subscription Renewal Dates for All Active Subscribers endpoint.
        :param product_id:         The product identifier of the auto-renewable subscription that you request a renewal-date extension for.
        :return: A response that indicates the current status of a request to extend the subscription renewal date to all eligible subscribers.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/subscriptions/extend/mass/" + product_id + "/" + request_identifier, "GET", {}, None, MassExtendRenewalDateStatusResponse)
    
    def get_test_notification_status(self, test_notification_token: str) -> CheckTestNotificationResponse:
        """
        Check the status of the test App Store server notification sent to your server.
        https://developer.apple.com/documentation/appstoreserverapi/get_test_notification_status

        :param test_notification_token: The test notification token received from the Request a Test Notification endpoint
        :return: A response that contains the contents of the test notification sent by the App Store server and the result from your server.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/notifications/test/" + test_notification_token, "GET", {}, None, CheckTestNotificationResponse)
    
    def get_notification_history(self, pagination_token: str, notification_history_request: NotificationHistoryRequest) -> NotificationHistoryResponse:
        """
        Get a list of notifications that the App Store server attempted to send to your server.
        https://developer.apple.com/documentation/appstoreserverapi/get_notification_history

        :param pagination_token: An optional token you use to get the next set of up to 20 notification history records. All responses that have more records available include a paginationToken. Omit this parameter the first time you call this endpoint.
        :param notification_history_request: The request body that includes the start and end dates, and optional query constraints.
        :return: A response that contains the App Store Server Notifications history for your app.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        queryParameters: Dict[str, List[str]] = dict()
        if pagination_token != None:
            queryParameters["paginationToken"] = [pagination_token]
        
        return self._make_request("/inApps/v1/notifications/history", "POST", queryParameters, notification_history_request, NotificationHistoryResponse)

    def get_transaction_history(self, transaction_id: str, revision: Optional[str], transaction_history_request: TransactionHistoryRequest) -> HistoryResponse:
        """
        Get a customer's in-app purchase transaction history for your app.
        https://developer.apple.com/documentation/appstoreserverapi/get_transaction_history

        :param transaction_id: The identifier of a transaction that belongs to the customer, and which may be an original transaction identifier.
        :param revision:              A token you provide to get the next set of up to 20 transactions. All responses include a revision token. Note: For requests that use the revision token, include the same query parameters from the initial request. Use the revision token from the previous HistoryResponse.
        :return: A response that contains the customer's transaction history for an app.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        queryParameters: Dict[str, List[str]] = dict()
        if revision != None:
            queryParameters["revision"] = [revision]
        
        if transaction_history_request.startDate != None:
            queryParameters["startDate"] = [str(transaction_history_request.startDate)]
        
        if transaction_history_request.endDate != None:
            queryParameters["endDate"] = [str(transaction_history_request.endDate)]
        
        if transaction_history_request.productIds != None:
            queryParameters["productId"] = transaction_history_request.productIds
        
        if transaction_history_request.productTypes != None:
            queryParameters["productType"] = [product_type.value for product_type in transaction_history_request.productTypes]
        
        if transaction_history_request.sort != None:
            queryParameters["sort"] = [transaction_history_request.sort.value]
        
        if transaction_history_request.subscriptionGroupIdentifiers != None:
            queryParameters["subscriptionGroupIdentifier"] = transaction_history_request.subscriptionGroupIdentifiers
        
        if transaction_history_request.inAppOwnershipType != None:
            queryParameters["inAppOwnershipType"] = [transaction_history_request.inAppOwnershipType.value]
        
        if transaction_history_request.revoked != None:
            queryParameters["revoked"] = [str(transaction_history_request.revoked)]
        
        return self._make_request("/inApps/v1/history/" + transaction_id, "GET", queryParameters, None, HistoryResponse)
    
    def get_transaction_info(self, transaction_id: str) -> TransactionInfoResponse:
        """
        Get information about a single transaction for your app.
        https://developer.apple.com/documentation/appstoreserverapi/get_transaction_info
        
        :param transaction_id The identifier of a transaction that belongs to the customer, and which may be an original transaction identifier.
        :return: A response that contains signed transaction information for a single transaction.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/transactions/" + transaction_id, "GET", {}, None, TransactionInfoResponse)

    def look_up_order_id(self, order_id: str) -> OrderLookupResponse:
        """
        Get a customer's in-app purchases from a receipt using the order ID.
        https://developer.apple.com/documentation/appstoreserverapi/look_up_order_id
        
        :param order_id: The order ID for in-app purchases that belong to the customer.
        :return: A response that includes the order lookup status and an array of signed transactions for the in-app purchases in the order.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/lookup/" + order_id, "GET", {}, None, OrderLookupResponse)
    def request_test_notification(self) -> SendTestNotificationResponse:
        """
        Ask App Store Server Notifications to send a test notification to your server.
        https://developer.apple.com/documentation/appstoreserverapi/request_a_test_notification

        :return: A response that contains the test notification token.
        :throws APIException: If a response was returned indicating the request could not be processed
        """
        return self._make_request("/inApps/v1/notifications/test", "POST", {}, None, SendTestNotificationResponse)

    def send_consumption_data(self, transaction_id: str, consumption_request: ConsumptionRequest):
        """
        Send consumption information about a consumable in-app purchase to the App Store after your server receives a consumption request notification.
        https://developer.apple.com/documentation/appstoreserverapi/send_consumption_information

        :param transaction_id: The transaction identifier for which you're providing consumption information. You receive this identifier in the CONSUMPTION_REQUEST notification the App Store sends to your server.
        :param consumption_request:    The request body containing consumption information.
        :raises APIException: If a response was returned indicating the request could not be processed
        """
        self._make_request("/inApps/v1/transactions/consumption/" + transaction_id, "PUT", {}, consumption_request, None)



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr
from .AutoRenewStatus import AutoRenewStatus
from .Environment import Environment

from .ExpirationIntent import ExpirationIntent
from .LibraryUtility import AttrsRawValueAware
from .OfferType import OfferType
from .PriceIncreaseStatus import PriceIncreaseStatus

@define
class JWSRenewalInfoDecodedPayload(AttrsRawValueAware):
    """
    A decoded payload containing subscription renewal information for an auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/jwsrenewalinfodecodedpayload
    """

    expirationIntent: Optional[ExpirationIntent] = ExpirationIntent.create_main_attr('rawExpirationIntent')
    """
    The reason the subscription expired.
    
    https://developer.apple.com/documentation/appstoreserverapi/expirationintent
    """

    rawExpirationIntent: Optional[int] = ExpirationIntent.create_raw_attr('expirationIntent')
    """
    See expirationIntent
    """

    originalTransactionId: Optional[str] = attr.ib(default=None)
    """
    The original transaction identifier of a purchase.
    
    https://developer.apple.com/documentation/appstoreserverapi/originaltransactionid
    """

    autoRenewProductId: Optional[str] = attr.ib(default=None)
    """
    The product identifier of the product that will renew at the next billing period.
    
    https://developer.apple.com/documentation/appstoreserverapi/autorenewproductid
    """

    productId: Optional[str] = attr.ib(default=None)
    """
    The unique identifier for the product, that you create in App Store Connect.
    
    https://developer.apple.com/documentation/appstoreserverapi/productid
    """

    autoRenewStatus: Optional[AutoRenewStatus] = AutoRenewStatus.create_main_attr('rawAutoRenewStatus')
    """
    The renewal status of the auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/autorenewstatus
    """

    rawAutoRenewStatus: Optional[int] = AutoRenewStatus.create_raw_attr('autoRenewStatus')
    """
    See autoRenewStatus
    """

    isInBillingRetryPeriod: Optional[bool] = attr.ib(default=None)
    """
    A Boolean value that indicates whether the App Store is attempting to automatically renew an expired subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/isinbillingretryperiod
    """

    priceIncreaseStatus: Optional[PriceIncreaseStatus] = PriceIncreaseStatus.create_main_attr('rawPriceIncreaseStatus')
    """
    The status that indicates whether the auto-renewable subscription is subject to a price increase.
    
    https://developer.apple.com/documentation/appstoreserverapi/priceincreasestatus
    """

    rawPriceIncreaseStatus: Optional[int] = PriceIncreaseStatus.create_raw_attr('priceIncreaseStatus')
    """
    See priceIncreaseStatus
    """

    gracePeriodExpiresDate: Optional[int] = attr.ib(default=None)
    """
    The time when the billing grace period for subscription renewals expires.
    
    https://developer.apple.com/documentation/appstoreserverapi/graceperiodexpiresdate
    """

    offerType: Optional[OfferType] = OfferType.create_main_attr('rawOfferType')
    """
    The type of the subscription offer.
    
    https://developer.apple.com/documentation/appstoreserverapi/offertype
    """

    rawOfferType: Optional[int] = OfferType.create_raw_attr('offerType')
    """
    See offerType
    """

    offerIdentifier: Optional[str] = attr.ib(default=None)
    """
    The identifier that contains the promo code or the promotional offer identifier.
    
    https://developer.apple.com/documentation/appstoreserverapi/offeridentifier
    """

    signedDate: Optional[int] = attr.ib(default=None)
    """
    The UNIX time, in milliseconds, that the App Store signed the JSON Web Signature data.
    
    https://developer.apple.com/documentation/appstoreserverapi/signeddate
    """

    environment: Optional[Environment] = Environment.create_main_attr('rawEnvironment')
    """
    The server environment, either sandbox or production.
    
    https://developer.apple.com/documentation/appstoreserverapi/environment
    """

    rawEnvironment: Optional[str] = Environment.create_raw_attr('environment')
    """
    See environment
    """

    recentSubscriptionStartDate: Optional[int] = attr.ib(default=None)
    """
    The earliest start date of a subscription in a series of auto-renewable subscription purchases that ignores all lapses of paid service shorter than 60 days.
    
    https://developer.apple.com/documentation/appstoreserverapi/recentsubscriptionstartdate
    """

    renewalDate: Optional[int] = attr.ib(default=None)
    """
    The UNIX time, in milliseconds, that the most recent auto-renewable subscription purchase expires.
    
    https://developer.apple.com/documentation/appstoreserverapi/renewaldate
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from base64 import b64decode
from typing import Optional

import asn1
import base64
import re

PKCS7_OID = "1.2.840.113549.1.7.2"
IN_APP_ARRAY = 17
TRANSACTION_IDENTIFIER = 1703

class ReceiptUtility:
    def extract_transaction_id_from_app_receipt(self, app_receipt: str) -> Optional[str]:
        """
        Extracts a transaction id from an encoded App Receipt. Throws if the receipt does not match the expected format.
        *NO validation* is performed on the receipt, and any data returned should only be used to call the App Store Server API.

        :param appReceipt: The unmodified app receipt
        :return: A transaction id from the array of in-app purchases, null if the receipt contains no in-app purchases
        """
        decoder = IndefiniteFormAwareDecoder()
        decoder.start(b64decode(app_receipt, validate=True))
        tag = decoder.peek()
        if tag.typ != asn1.Types.Constructed or tag.nr != asn1.Numbers.Sequence:
            raise ValueError()
        decoder.enter()
        # PKCS#7 object
        tag, value = decoder.read()
        if tag.typ != asn1.Types.Primitive or tag.nr != asn1.Numbers.ObjectIdentifier or value != PKCS7_OID:
            raise ValueError()
        # This is the PKCS#7 format, work our way into the inner content
        decoder.enter()
        decoder.enter()
        decoder.read()
        decoder.read()
        decoder.enter()
        decoder.read()
        decoder.enter()
        tag, value = decoder.read()
        # Xcode uses nested OctetStrings, we extract the inner string in this case
        if tag.typ == asn1.Types.Constructed and tag.nr == asn1.Numbers.OctetString:
            inner_decoder = asn1.Decoder()
            inner_decoder.start(value)
            tag, value = inner_decoder.read()
        if tag.typ != asn1.Types.Primitive or tag.nr != asn1.Numbers.OctetString:
            raise ValueError()
        decoder = asn1.Decoder()
        decoder.start(value)
        tag = decoder.peek()
        if tag.typ != asn1.Types.Constructed or tag.nr != asn1.Numbers.Set:
            raise ValueError()
        decoder.enter()
        # We are in the top-level sequence, work our way to the array of in-apps
        while not decoder.eof():
            decoder.enter()
            tag, value = decoder.read()
            if tag.typ == asn1.Types.Primitive and tag.nr == asn1.Numbers.Integer and value == IN_APP_ARRAY:
                decoder.read()
                tag, value = decoder.read()
                if tag.typ != asn1.Types.Primitive or tag.nr != asn1.Numbers.OctetString:
                    raise ValueError()
                inapp_decoder = asn1.Decoder()
                inapp_decoder.start(value)
                inapp_decoder.enter()
                # In-app array
                while not inapp_decoder.eof():
                    inapp_decoder.enter()
                    tag, value = inapp_decoder.read()
                    if (
                        tag.typ == asn1.Types.Primitive
                        and tag.nr == asn1.Numbers.Integer
                        and value == TRANSACTION_IDENTIFIER
                    ):
                        inapp_decoder.read()
                        tag, value = inapp_decoder.read()
                        singleton_decoder = asn1.Decoder()
                        singleton_decoder.start(value)
                        tag, value = singleton_decoder.read()
                        return value
                    inapp_decoder.leave()
            decoder.leave()
        return None
    
    def extract_transaction_id_from_transaction_receipt(self, transaction_receipt: str) -> Optional[str]:
        """
        Extracts a transaction id from an encoded transactional receipt. Throws if the receipt does not match the expected format.
        *NO validation* is performed on the receipt, and any data returned should only be used to call the App Store Server API.
        :param transactionReceipt: The unmodified transactionReceipt
        :return: A transaction id, or null if no transactionId is found in the receipt
        """
        decoded_top_level = base64.b64decode(transaction_receipt).decode('utf-8')
        matching_result = re.search('"purchase-info"\s+=\s+"([a-zA-Z0-9+/=]+)";', decoded_top_level)
        if matching_result:
            decoded_inner_level = base64.b64decode(matching_result.group(1)).decode('utf-8')
            inner_matching_result = re.search('"transaction-id"\s+=\s+"([a-zA-Z0-9+/=]+)";', decoded_inner_level)
            if inner_matching_result:
                return inner_matching_result.group(1)
        return None

class IndefiniteFormAwareDecoder(asn1.Decoder):
    def _read_length(self) -> int:
        index, input_data = self.m_stack[-1]
        try:
            byte = input_data[index]
        except IndexError:
            raise asn1.Error('Premature end of input.')
        if byte == 0x80:
            # Xcode receipts use indefinite length encoding, not supported by all parsers
            # Indefinite length encoding is only entered, but never left during parsing for receipts
            # We therefore round up indefinite length encoding to be the remaining length
            self._read_byte()
            index, input_data = self.m_stack[-1]
            return len(input_data) - index
        return super()._read_length()


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

@define
class SendTestNotificationResponse: 
    """
    A response that contains the test notification token.
    
    https://developer.apple.com/documentation/appstoreserverapi/sendtestnotificationresponse
    """

    testNotificationToken: Optional[str] = attr.ib(default=None)
    """
    A unique identifier for a notification test that the App Store server sends to your server.
    
    https://developer.apple.com/documentation/appstoreserverapi/testnotificationtoken
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class OfferDiscountType(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The payment mode you configure for an introductory offer, promotional offer, or offer code on an auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/offerdiscounttype
    """
    FREE_TRIAL = "FREE_TRIAL"
    PAY_AS_YOU_GO = "PAY_AS_YOU_GO"
    PAY_UP_FRONT = "PAY_UP_FRONT"


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from typing import Any, Dict, List, Union
import unittest

from requests import Response
from appstoreserverlibrary.api_client import APIError, APIException, AppStoreServerAPIClient
from appstoreserverlibrary.models.AccountTenure import AccountTenure
from appstoreserverlibrary.models.AutoRenewStatus import AutoRenewStatus
from appstoreserverlibrary.models.ConsumptionRequest import ConsumptionRequest
from appstoreserverlibrary.models.ConsumptionStatus import ConsumptionStatus
from appstoreserverlibrary.models.DeliveryStatus import DeliveryStatus
from appstoreserverlibrary.models.Environment import Environment
from appstoreserverlibrary.models.ExpirationIntent import ExpirationIntent
from appstoreserverlibrary.models.ExtendReasonCode import ExtendReasonCode
from appstoreserverlibrary.models.ExtendRenewalDateRequest import ExtendRenewalDateRequest
from appstoreserverlibrary.models.InAppOwnershipType import InAppOwnershipType
from appstoreserverlibrary.models.LastTransactionsItem import LastTransactionsItem
from appstoreserverlibrary.models.LifetimeDollarsPurchased import LifetimeDollarsPurchased
from appstoreserverlibrary.models.LifetimeDollarsRefunded import LifetimeDollarsRefunded
from appstoreserverlibrary.models.MassExtendRenewalDateRequest import MassExtendRenewalDateRequest
from appstoreserverlibrary.models.NotificationHistoryRequest import NotificationHistoryRequest
from appstoreserverlibrary.models.NotificationHistoryResponseItem import NotificationHistoryResponseItem
from appstoreserverlibrary.models.NotificationTypeV2 import NotificationTypeV2
from appstoreserverlibrary.models.OfferType import OfferType
from appstoreserverlibrary.models.OrderLookupStatus import OrderLookupStatus
from appstoreserverlibrary.models.Platform import Platform
from appstoreserverlibrary.models.PlayTime import PlayTime
from appstoreserverlibrary.models.PriceIncreaseStatus import PriceIncreaseStatus
from appstoreserverlibrary.models.RevocationReason import RevocationReason
from appstoreserverlibrary.models.SendAttemptItem import SendAttemptItem
from appstoreserverlibrary.models.SendAttemptResult import SendAttemptResult
from appstoreserverlibrary.models.Status import Status
from appstoreserverlibrary.models.SubscriptionGroupIdentifierItem import SubscriptionGroupIdentifierItem
from appstoreserverlibrary.models.Subtype import Subtype
from appstoreserverlibrary.models.TransactionHistoryRequest import Order, ProductType, TransactionHistoryRequest
from appstoreserverlibrary.models.TransactionReason import TransactionReason
from appstoreserverlibrary.models.Type import Type
from appstoreserverlibrary.models.UserStatus import UserStatus

from tests.util import decode_json_from_signed_date, read_data_from_binary_file, read_data_from_file

from io import BytesIO

class DecodedPayloads(unittest.TestCase):
    def test_extend_renewal_date_for_all_active_subscribers(self):
        client = self.get_client_with_body_from_file('tests/resources/models/extendRenewalDateForAllActiveSubscribersResponse.json',
                                           'POST',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/subscriptions/extend/mass',
                                           {}, 
                                           {'extendByDays': 45, 'extendReasonCode': 1, 'requestIdentifier': 'fdf964a4-233b-486c-aac1-97d8d52688ac', 'storefrontCountryCodes': ['USA', 'MEX'], 'productId': 'com.example.productId'})
        
        extend_renewal_date_request = MassExtendRenewalDateRequest(
             extendByDays=45,
             extendReasonCode=ExtendReasonCode.CUSTOMER_SATISFACTION,
             requestIdentifier='fdf964a4-233b-486c-aac1-97d8d52688ac',
             storefrontCountryCodes=['USA', 'MEX'],
             productId='com.example.productId')

        mass_extend_renewal_date_response = client.extend_renewal_date_for_all_active_subscribers(extend_renewal_date_request)

        self.assertIsNotNone(mass_extend_renewal_date_response)
        self.assertEqual('758883e8-151b-47b7-abd0-60c4d804c2f5', mass_extend_renewal_date_response.requestIdentifier)

    def test_extend_subscription_renewal_date(self):
        client = self.get_client_with_body_from_file('tests/resources/models/extendSubscriptionRenewalDateResponse.json',
                                           'PUT',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/subscriptions/extend/4124214',
                                           {},
                                           {'extendByDays': 45, 'extendReasonCode': 1, 'requestIdentifier': 'fdf964a4-233b-486c-aac1-97d8d52688ac'})

        extend_renewal_date_request = ExtendRenewalDateRequest(
             extendByDays=45,
             extendReasonCode=ExtendReasonCode.CUSTOMER_SATISFACTION,
             requestIdentifier='fdf964a4-233b-486c-aac1-97d8d52688ac'
        )

        extend_renewal_date_response = client.extend_subscription_renewal_date('4124214', extend_renewal_date_request)

        self.assertIsNotNone(extend_renewal_date_response)
        self.assertEqual('2312412', extend_renewal_date_response.originalTransactionId)
        self.assertEqual('9993', extend_renewal_date_response.webOrderLineItemId)
        self.assertTrue(extend_renewal_date_response.success)
        self.assertEqual(1698148900000, extend_renewal_date_response.effectiveDate)

    def test_get_all_subscription_statuses(self):
        client = self.get_client_with_body_from_file('tests/resources/models/getAllSubscriptionStatusesResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/subscriptions/4321', 
                                           {'status': [2, 1]},
                                           None)

        status_response = client.get_all_subscription_statuses('4321', [Status.EXPIRED, Status.ACTIVE])

        self.assertIsNotNone(status_response)
        self.assertEqual(Environment.LOCAL_TESTING, status_response.environment)
        self.assertEqual('LocalTesting', status_response.rawEnvironment)
        self.assertEqual('com.example', status_response.bundleId)
        self.assertEqual(5454545, status_response.appAppleId)


        expected_body = [
             SubscriptionGroupIdentifierItem(
                subscriptionGroupIdentifier='sub_group_one',
                lastTransactions=[
                    LastTransactionsItem(
                        status=Status.ACTIVE,
                        originalTransactionId='3749183',
                        signedTransactionInfo='signed_transaction_one',
                        signedRenewalInfo='signed_renewal_one'
                    ),
                    LastTransactionsItem(
                        status=Status.REVOKED,
                        originalTransactionId='5314314134',
                        signedTransactionInfo='signed_transaction_two',
                        signedRenewalInfo='signed_renewal_two'
                    )
                ]
            ),
            SubscriptionGroupIdentifierItem(
                 subscriptionGroupIdentifier='sub_group_two',
                 lastTransactions=[
                      LastTransactionsItem(
                           status=Status.EXPIRED,
                           originalTransactionId='3413453',
                           signedTransactionInfo='signed_transaction_three',
                           signedRenewalInfo='signed_renewal_three'
                      )
                 ]
            )
        ]
        self.assertEqual(expected_body, status_response.data)

    def test_get_refund_history(self):
        client = self.get_client_with_body_from_file('tests/resources/models/getRefundHistoryResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v2/refund/lookup/555555', 
                                           {'revision': ['revision_input']}, 
                                           None)

        refund_history_response = client.get_refund_history('555555', 'revision_input')

        self.assertIsNotNone(refund_history_response)
        self.assertEqual(['signed_transaction_one', 'signed_transaction_two'], refund_history_response.signedTransactions)
        self.assertEqual('revision_output', refund_history_response.revision)
        self.assertTrue(refund_history_response.hasMore)

    def test_get_status_of_subscription_renewal_date_extensions(self):
        client = self.get_client_with_body_from_file('tests/resources/models/getStatusOfSubscriptionRenewalDateExtensionsResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/subscriptions/extend/mass/20fba8a0-2b80-4a7d-a17f-85c1854727f8/com.example.product', 
                                           {},
                                           None)

        mass_extend_renewal_date_status_response = client.get_status_of_subscription_renewal_date_extensions('com.example.product', '20fba8a0-2b80-4a7d-a17f-85c1854727f8')

        self.assertIsNotNone(mass_extend_renewal_date_status_response)
        self.assertEqual('20fba8a0-2b80-4a7d-a17f-85c1854727f8', mass_extend_renewal_date_status_response.requestIdentifier)
        self.assertTrue(mass_extend_renewal_date_status_response.complete)
        self.assertEqual(1698148900000, mass_extend_renewal_date_status_response.completeDate)
        self.assertEqual(30, mass_extend_renewal_date_status_response.succeededCount)
        self.assertEqual(2, mass_extend_renewal_date_status_response.failedCount)

    def test_get_test_notification_status(self):
        client = self.get_client_with_body_from_file('tests/resources/models/getTestNotificationStatusResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/notifications/test/8cd2974c-f905-492a-bf9a-b2f47c791d19',
                                           {},
                                           None)

        check_test_notification_response = client.get_test_notification_status('8cd2974c-f905-492a-bf9a-b2f47c791d19')

        self.assertIsNotNone(check_test_notification_response)
        self.assertEqual('signed_payload', check_test_notification_response.signedPayload)
        sendAttemptItems = [
                                SendAttemptItem(attemptDate=1698148900000,sendAttemptResult=SendAttemptResult.NO_RESPONSE),
                                SendAttemptItem(attemptDate=1698148950000,sendAttemptResult=SendAttemptResult.SUCCESS)
                            ]
        self.assertEqual(sendAttemptItems, check_test_notification_response.sendAttempts)

    def test_get_notification_history(self):
        client = self.get_client_with_body_from_file('tests/resources/models/getNotificationHistoryResponse.json',
                                           'POST',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/notifications/history', 
                                           {'paginationToken': ['a036bc0e-52b8-4bee-82fc-8c24cb6715d6']},
                                           {'startDate': 1698148900000, 'endDate': 1698148950000, 'notificationType': 'SUBSCRIBED', 'notificationSubtype': 'INITIAL_BUY', 'transactionId': '999733843', 'onlyFailures': True})

        notification_history_request = NotificationHistoryRequest(
            startDate=1698148900000,
            endDate=1698148950000,
            notificationType=NotificationTypeV2.SUBSCRIBED,
            notificationSubtype=Subtype.INITIAL_BUY,
            transactionId='999733843',
            onlyFailures=True
        )
        
        notification_history_response = client.get_notification_history('a036bc0e-52b8-4bee-82fc-8c24cb6715d6', notification_history_request)

        self.assertIsNotNone(notification_history_response)
        self.assertEqual('57715481-805a-4283-8499-1c19b5d6b20a', notification_history_response.paginationToken)
        self.assertTrue(notification_history_response.hasMore)
        expected_notification_history = [
             NotificationHistoryResponseItem(sendAttempts=[
                  SendAttemptItem(
                       attemptDate=1698148900000,
                       sendAttemptResult=SendAttemptResult.NO_RESPONSE
                  ),
                  SendAttemptItem(
                       attemptDate=1698148950000,
                       rawSendAttemptResult='SUCCESS'
                  )
             ], signedPayload='signed_payload_one'),
             NotificationHistoryResponseItem(sendAttempts=[
                  SendAttemptItem(
                       attemptDate=1698148800000,
                        sendAttemptResult=SendAttemptResult.CIRCULAR_REDIRECT
                    )
             ], signedPayload='signed_payload_two')
        ]
        self.assertEqual(expected_notification_history, notification_history_response.notificationHistory)

    def test_get_transaction_history(self):
        client = self.get_client_with_body_from_file('tests/resources/models/transactionHistoryResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/history/1234', 
                                           {'revision': ['revision_input'],
                                            'startDate': ['123455'],
                                            'endDate': ['123456'],
                                            'productId': ['com.example.1', 'com.example.2'],
                                            'productType': ['CONSUMABLE', 'AUTO_RENEWABLE'],
                                            'sort': ['ASCENDING'],
                                            'subscriptionGroupIdentifier': ['sub_group_id', 'sub_group_id_2'],
                                            'inAppOwnershipType': ['FAMILY_SHARED'],
                                            'revoked': ['False']},
                                            None)

        request = TransactionHistoryRequest(
            sort=Order.ASCENDING,
            productTypes=[ProductType.CONSUMABLE, ProductType.AUTO_RENEWABLE],
            endDate=123456,
            startDate=123455,
            revoked=False,
            inAppOwnershipType=InAppOwnershipType.FAMILY_SHARED,
            productIds=['com.example.1', 'com.example.2'],
            subscriptionGroupIdentifiers=['sub_group_id', 'sub_group_id_2']
        )

        history_response = client.get_transaction_history('1234', 'revision_input', request)

        self.assertIsNotNone(history_response)
        self.assertEqual('revision_output', history_response.revision)
        self.assertTrue(history_response.hasMore)
        self.assertEqual('com.example', history_response.bundleId)
        self.assertEqual(323232, history_response.appAppleId)
        self.assertEqual(Environment.LOCAL_TESTING, history_response.environment)
        self.assertEqual('LocalTesting', history_response.rawEnvironment)
        self.assertEqual(['signed_transaction_value', 'signed_transaction_value2'], history_response.signedTransactions)

    def test_get_transaction_info(self):
        client = self.get_client_with_body_from_file('tests/resources/models/transactionInfoResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/transactions/1234', 
                                           {},
                                           None)

        transaction_info_response = client.get_transaction_info('1234')

        self.assertIsNotNone(transaction_info_response)
        self.assertEqual('signed_transaction_info_value', transaction_info_response.signedTransactionInfo)

    def test_look_up_order_id(self):
        client = self.get_client_with_body_from_file('tests/resources/models/lookupOrderIdResponse.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/lookup/W002182',
                                           {},
                                           None)

        order_lookup_response = client.look_up_order_id('W002182')

        self.assertIsNotNone(order_lookup_response)
        self.assertEqual(OrderLookupStatus.INVALID, order_lookup_response.status)
        self.assertEqual(1, order_lookup_response.rawStatus)
        self.assertEqual(['signed_transaction_one', 'signed_transaction_two'], order_lookup_response.signedTransactions)

    def test_request_test_notification(self):
        client = self.get_client_with_body_from_file('tests/resources/models/requestTestNotificationResponse.json',
                                           'POST',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/notifications/test',
                                           {},
                                           None)

        send_test_notification_response = client.request_test_notification()

        self.assertIsNotNone(send_test_notification_response)
        self.assertEqual('ce3af791-365e-4c60-841b-1674b43c1609', send_test_notification_response.testNotificationToken)

    def test_send_consumption_data(self):
        client = self.get_client_with_body(b'',
                                           'PUT',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/transactions/consumption/49571273', 
                                           {},
                                           {'customerConsented': True,
                                            'consumptionStatus': 1,
                                            'platform': 2,
                                            'sampleContentProvided': False,
                                            'deliveryStatus': 3,
                                            'appAccountToken': '7389a31a-fb6d-4569-a2a6-db7d85d84813',
                                            'accountTenure': 4,
                                            'playTime': 5,
                                            'lifetimeDollarsRefunded': 6,
                                            'lifetimeDollarsPurchased': 7,
                                            'userStatus': 4})

        consumptionRequest = ConsumptionRequest(
            customerConsented=True,
            consumptionStatus=ConsumptionStatus.NOT_CONSUMED,
            platform=Platform.NON_APPLE,
            sampleContentProvided=False,
            deliveryStatus=DeliveryStatus.DID_NOT_DELIVER_DUE_TO_SERVER_OUTAGE,
            appAccountToken='7389a31a-fb6d-4569-a2a6-db7d85d84813',
            accountTenure=AccountTenure.THIRTY_DAYS_TO_NINETY_DAYS,
            playTime=PlayTime.ONE_DAY_TO_FOUR_DAYS,
            lifetimeDollarsRefunded=LifetimeDollarsRefunded.ONE_THOUSAND_DOLLARS_TO_ONE_THOUSAND_NINE_HUNDRED_NINETY_NINE_DOLLARS_AND_NINETY_NINE_CENTS,
            lifetimeDollarsPurchased=LifetimeDollarsPurchased.TWO_THOUSAND_DOLLARS_OR_GREATER,
            userStatus=UserStatus.LIMITED_ACCESS
        )

        client.send_consumption_data('49571273', consumptionRequest)

    def test_api_error(self):
        client = self.get_client_with_body_from_file('tests/resources/models/apiException.json',
                                                     'POST',
                                                     'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/notifications/test',
                                                     {},
                                                     None,
                                                     500)
        try:
            client.request_test_notification()
        except APIException as e:
            self.assertEqual(500, e.http_status_code)
            self.assertEqual(5000000, e.raw_api_error)
            self.assertEqual(APIError.GENERAL_INTERNAL, e.api_error)
            self.assertEqual("An unknown error occurred.", e.error_message)
            return
        
        self.assertFalse(True)
    
    def test_api_too_many_requests(self):
        client = self.get_client_with_body_from_file('tests/resources/models/apiTooManyRequestsException.json',
                                                     'POST',
                                                     'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/notifications/test',
                                                     {},
                                                     None,
                                                     429)
        try:
            client.request_test_notification()
        except APIException as e:
            self.assertEqual(429, e.http_status_code)
            self.assertEqual(4290000, e.raw_api_error)
            self.assertEqual(APIError.RATE_LIMIT_EXCEEDED, e.api_error)
            self.assertEqual("Rate limit exceeded.", e.error_message)
            return
        
        self.assertFalse(True)

    def test_unknown_error(self):
        client = self.get_client_with_body_from_file('tests/resources/models/apiUnknownError.json',
                                                     'POST',
                                                     'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/notifications/test',
                                                     {},
                                                     None,
                                                     400)
        try:
            client.request_test_notification()
        except APIException as e:
            self.assertEqual(400, e.http_status_code)
            self.assertEqual(9990000, e.raw_api_error)
            self.assertIsNone(e.api_error)
            self.assertEqual("Testing error.", e.error_message)
            return
        
        self.assertFalse(True)

    def test_get_transaction_history_with_unknown_environment(self):
        client = self.get_client_with_body_from_file('tests/resources/models/transactionHistoryResponseWithMalformedEnvironment.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/history/1234', 
                                           {'revision': ['revision_input'],
                                            'startDate': ['123455'],
                                            'endDate': ['123456'],
                                            'productId': ['com.example.1', 'com.example.2'],
                                            'productType': ['CONSUMABLE', 'AUTO_RENEWABLE'],
                                            'sort': ['ASCENDING'],
                                            'subscriptionGroupIdentifier': ['sub_group_id', 'sub_group_id_2'],
                                            'inAppOwnershipType': ['FAMILY_SHARED'],
                                            'revoked': ['False']},
                                            None)

        request = TransactionHistoryRequest(
            sort=Order.ASCENDING,
            productTypes=[ProductType.CONSUMABLE, ProductType.AUTO_RENEWABLE],
            endDate=123456,
            startDate=123455,
            revoked=False,
            inAppOwnershipType=InAppOwnershipType.FAMILY_SHARED,
            productIds=['com.example.1', 'com.example.2'],
            subscriptionGroupIdentifiers=['sub_group_id', 'sub_group_id_2']
        )

        history_response = client.get_transaction_history('1234', 'revision_input', request)

        self.assertIsNone(history_response.environment)
        self.assertEqual("LocalTestingxxx", history_response.rawEnvironment)

    def test_get_transaction_history_with_malformed_app_apple_id(self):
        client = self.get_client_with_body_from_file('tests/resources/models/transactionHistoryResponseWithMalformedAppAppleId.json',
                                           'GET',
                                           'https://api.storekit-sandbox.itunes.apple.com/inApps/v1/history/1234', 
                                           {'revision': ['revision_input'],
                                            'startDate': ['123455'],
                                            'endDate': ['123456'],
                                            'productId': ['com.example.1', 'com.example.2'],
                                            'productType': ['CONSUMABLE', 'AUTO_RENEWABLE'],
                                            'sort': ['ASCENDING'],
                                            'subscriptionGroupIdentifier': ['sub_group_id', 'sub_group_id_2'],
                                            'inAppOwnershipType': ['FAMILY_SHARED'],
                                            'revoked': ['False']},
                                            None)

        request = TransactionHistoryRequest(
            sort=Order.ASCENDING,
            productTypes=[ProductType.CONSUMABLE, ProductType.AUTO_RENEWABLE],
            endDate=123456,
            startDate=123455,
            revoked=False,
            inAppOwnershipType=InAppOwnershipType.FAMILY_SHARED,
            productIds=['com.example.1', 'com.example.2'],
            subscriptionGroupIdentifiers=['sub_group_id', 'sub_group_id_2']
        )

        try:
            client.get_transaction_history('1234', 'revision_input', request)
        except Exception:
            return
        
        self.assertFalse(True)

    
    def get_client_with_body(self, body: str, expected_method: str, expected_url: str, expected_params: Dict[str, Union[str, List[str]]], expected_json: Dict[str, Any], status_code: int = 200):
        signing_key = read_data_from_binary_file('tests/resources/certs/testSigningKey.p8')
        client = AppStoreServerAPIClient(signing_key, 'keyId', 'issuerId', 'com.example', Environment.LOCAL_TESTING)
        def fake_execute_and_validate_inputs(method: bytes, url: str, params: Dict[str, Union[str, List[str]]], headers: Dict[str, str], json: Dict[str, Any]):
            self.assertEqual(expected_method, method)
            self.assertEqual(expected_url, url)
            self.assertEqual(expected_params, params)
            self.assertEqual(['User-Agent', 'Authorization', 'Accept'], list(headers.keys()))
            self.assertEqual('application/json', headers['Accept'])
            self.assertTrue(headers['User-Agent'].startswith('app-store-server-library/python'))
            self.assertTrue(headers['Authorization'].startswith('Bearer '))
            decoded_jwt = decode_json_from_signed_date(headers['Authorization'][7:])
            self.assertEqual('appstoreconnect-v1', decoded_jwt['payload']['aud'])
            self.assertEqual('issuerId', decoded_jwt['payload']['iss'])
            self.assertEqual('keyId', decoded_jwt['header']['kid'])
            self.assertEqual('com.example', decoded_jwt['payload']['bid'])
            self.assertEqual(expected_json, json)
            response = Response()
            response.status_code = status_code
            response.raw = BytesIO(body)
            response.headers['Content-Type'] = 'application/json'
            return response

        client._execute_request = fake_execute_and_validate_inputs
        return client

    def get_client_with_body_from_file(self, path: str, expected_method: str, expected_url: str, expected_params: Dict[str, Union[str, List[str]]], expected_json: Dict[str, Any], status_code: int = 200):
        body = read_data_from_binary_file(path)
        return self.get_client_with_body(body, expected_method, expected_url, expected_params, expected_json, status_code)



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class PriceIncreaseStatus(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The status that indicates whether an auto-renewable subscription is subject to a price increase.
    
    https://developer.apple.com/documentation/appstoreserverapi/priceincreasestatus
    """
    CUSTOMER_HAS_NOT_RESPONDED = 0
    CUSTOMER_CONSENTED_OR_WAS_NOTIFIED_WITHOUT_NEEDING_CONSENT = 1



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional, List

from attr import define
import attr

from .Environment import Environment
from .LibraryUtility import AttrsRawValueAware
from .SubscriptionGroupIdentifierItem import SubscriptionGroupIdentifierItem

@define
class StatusResponse(AttrsRawValueAware):
    """
    A response that contains status information for all of a customer's auto-renewable subscriptions in your app.
    
    https://developer.apple.com/documentation/appstoreserverapi/statusresponse
    """

    environment: Optional[Environment] = Environment.create_main_attr('rawEnvironment')
    """
    The server environment, sandbox or production, in which the App Store generated the response.
    
    https://developer.apple.com/documentation/appstoreserverapi/environment
    """
            
    rawEnvironment: Optional[str] = Environment.create_raw_attr('environment')
    """
    See environment
    """

    bundleId: Optional[str] = attr.ib(default=None)
    """
    The bundle identifier of an app.
    
    https://developer.apple.com/documentation/appstoreserverapi/bundleid
    """

    appAppleId: Optional[int] = attr.ib(default=None)
    """
    The unique identifier of an app in the App Store.
    
    https://developer.apple.com/documentation/appstoreservernotifications/appappleid
    """

    data: Optional[List[SubscriptionGroupIdentifierItem]] = attr.ib(default=None)
    """
    An array of information for auto-renewable subscriptions, including App Store-signed transaction information and App Store-signed renewal information.
    
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr
from .Data import Data

from .LibraryUtility import AttrsRawValueAware
from .NotificationTypeV2 import NotificationTypeV2
from .Subtype import Subtype
from .Summary import Summary


@define
class ResponseBodyV2DecodedPayload(AttrsRawValueAware):
    """
    A decoded payload containing the version 2 notification data.
    
    https://developer.apple.com/documentation/appstoreservernotifications/responsebodyv2decodedpayload
    """

    notificationType: Optional[NotificationTypeV2] = NotificationTypeV2.create_main_attr('rawNotificationType')
    """
    The in-app purchase event for which the App Store sends this version 2 notification.
    
    https://developer.apple.com/documentation/appstoreservernotifications/notificationtype
    """

    rawNotificationType: Optional[str] = NotificationTypeV2.create_raw_attr('notificationType')
    """
    See notificationType
    """

    subtype: Optional[Subtype] = Subtype.create_main_attr('rawSubtype')
    """
    Additional information that identifies the notification event. The subtype field is present only for specific version 2 notifications.
    
    https://developer.apple.com/documentation/appstoreservernotifications/subtype
    """
    
    rawSubtype: Optional[str] = Subtype.create_raw_attr('subtype')
    """
    See subtype
    """

    notificationUUID: Optional[str] = attr.ib(default=None)
    """
    A unique identifier for the notification.
    
    https://developer.apple.com/documentation/appstoreservernotifications/notificationuuid
    """

    data: Optional[Data] = attr.ib(default=None)
    """
    The object that contains the app metadata and signed renewal and transaction information.
    The data and summary fields are mutually exclusive. The payload contains one of the fields, but not both.
    
    https://developer.apple.com/documentation/appstoreservernotifications/data
    """

    version: Optional[str] = attr.ib(default=None)
    """
    A string that indicates the notification's App Store Server Notifications version number.
    
    https://developer.apple.com/documentation/appstoreservernotifications/version
    """

    signedDate: Optional[int] = attr.ib(default=None)
    """
    The UNIX time, in milliseconds, that the App Store signed the JSON Web Signature data.
    
    https://developer.apple.com/documentation/appstoreserverapi/signeddate
    """

    summary: Optional[Summary] = attr.ib(default=None)
    """
    The summary data that appears when the App Store server completes your request to extend a subscription renewal date for eligible subscribers.
    The data and summary fields are mutually exclusive. The payload contains one of the fields, but not both.
    
    https://developer.apple.com/documentation/appstoreservernotifications/summary
    """


# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class AccountTenure(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The age of the customer's account.
    
    https://developer.apple.com/documentation/appstoreserverapi/accounttenure
    """
    UNDECLARED = 0
    ZERO_TO_THREE_DAYS = 1
    THREE_DAYS_TO_TEN_DAYS = 2
    TEN_DAYS_TO_THIRTY_DAYS = 3
    THIRTY_DAYS_TO_NINETY_DAYS = 4
    NINETY_DAYS_TO_ONE_HUNDRED_EIGHTY_DAYS = 5
    ONE_HUNDRED_EIGHTY_DAYS_TO_THREE_HUNDRED_SIXTY_FIVE_DAYS = 6
    GREATER_THAN_THREE_HUNDRED_SIXTY_FIVE_DAYS = 7



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import Enum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class InAppOwnershipType(str, Enum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The relationship of the user with the family-shared purchase to which they have access.
    
    https://developer.apple.com/documentation/appstoreserverapi/inappownershiptype
    """
    FAMILY_SHARED = "FAMILY_SHARED"
    PURCHASED = "PURCHASED"



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.
from typing import Optional

from attr import define
import attr

@define
class TransactionInfoResponse:
    """
    A response that contains signed transaction information for a single transaction.
    
    https://developer.apple.com/documentation/appstoreserverapi/transactioninforesponse
    """
    
    signedTransactionInfo: Optional[str] = attr.ib(default=None)
    """
    A customer’s in-app purchase transaction, signed by Apple, in JSON Web Signature (JWS) format.
    
    https://developer.apple.com/documentation/appstoreserverapi/jwstransaction
    """



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class Status(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    The status of an auto-renewable subscription.
    
    https://developer.apple.com/documentation/appstoreserverapi/status
    """
    ACTIVE = 1
    EXPIRED = 2
    BILLING_RETRY = 3
    BILLING_GRACE_PERIOD = 4
    REVOKED = 5



# Copyright (c) 2023 Apple Inc. Licensed under MIT License.

from enum import IntEnum

from .LibraryUtility import AppStoreServerLibraryEnumMeta

class DeliveryStatus(IntEnum, metaclass=AppStoreServerLibraryEnumMeta):
    """
    A value that indicates whether the app successfully delivered an in-app purchase that works properly.
    
    https://developer.apple.com/documentation/appstoreserverapi/deliverystatus
    """
    DELIVERED_AND_WORKING_PROPERLY = 0
    DID_NOT_DELIVER_DUE_TO_QUALITY_ISSUE = 1
    DELIVERED_WRONG_ITEM = 2
    DID_NOT_DELIVER_DUE_TO_SERVER_OUTAGE = 3
    DID_NOT_DELIVER_DUE_TO_IN_GAME_CURRENCY_CHANGE = 4
    DID_NOT_DELIVER_FOR_OTHER_REASON = 5
